#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include "real.h"
#include "fft.h"
#include "functions.h"

struct POINTER_ITEM {
  struct POINTER_ITEM	*next;
  char			*p;
};

struct POINTER_LIST {
  struct POINTER_ITEM	*head;
  struct POINTER_ITEM	*tail;
  struct POINTER_ITEM	*cur;
  int			n_items;
};

void AddPointerItem(struct POINTER_LIST *, char *);
struct POINTER_LIST *NewPointerList(void);
void FreePointerList(struct POINTER_LIST *);
char ** PointerListToVector(struct POINTER_LIST *, int *);

struct REAL_ITEM {
  struct REAL_ITEM	*next;
  real			x;
};

struct REAL_LIST {
  struct REAL_ITEM	*head;
  struct REAL_ITEM	*tail;
  struct REAL_ITEM	*cur;
  int			n_items;
};

void AddRealItem(struct REAL_LIST *, real);
struct REAL_LIST *NewRealList(void);
void FreeRealList(struct REAL_LIST *);
real * RealListToVector(struct REAL_LIST *, int *);

void FreePointerList
  (struct POINTER_LIST *list)
{
  struct POINTER_ITEM *item, *next;
  if (list==NULL)
    return;
  item=list->head;
  while (item!=NULL) {
    next = item->next;
    free(item);
    item = next;
  }
  free(list);
}

void AddPointerItem
  (struct POINTER_LIST *list,
   char *x)
{
  struct POINTER_ITEM *item;
  item = CallocOrAbort(struct POINTER_ITEM, 1);
  item->p = x;
  item->next = NULL;
  if (list->head==NULL)
    list->head = item;
  else
    list->tail->next = item;
  list->tail = item;
  list->n_items++;
}

struct POINTER_LIST * NewPointerList(void)
{
  struct POINTER_LIST *list;
  list = CallocOrAbort(struct POINTER_LIST, 1);
  list->head = list->tail = list->cur = NULL;
  list->n_items = 0;
  return list;
}

char **PointerListToVector
  (struct POINTER_LIST *list,
   int *n)
{
  int i = 0;
  char **v = CallocOrAbort(char *, list->n_items);
  struct POINTER_ITEM *p;
  for (p=list->head; p!=NULL; p=p->next)
    v[i++] = p->p;
  if (n!=NULL)
    *n = list->n_items;
  return v;
}

void FreeRealList
  (struct REAL_LIST *list)
{
  struct REAL_ITEM *item, *next;
  if (list==NULL)
    return;
  item=list->head;
  while (item!=NULL) {
    next = item->next;
    free(item);
    item = next;
  }
  free(list);
}

void AddRealItem(struct REAL_LIST *list, real x)
{
  struct REAL_ITEM *item;
  item = CallocOrAbort(struct REAL_ITEM, 1);
  item->x = x;
  item->next = NULL;
  if (list->head==NULL)
    list->head = item;
  else
    list->tail->next = item;
  list->tail = item;
  list->n_items++;
}

struct REAL_LIST * NewRealList(void)
{
  struct REAL_LIST *list;
  list = CallocOrAbort(struct REAL_LIST, 1);
  list->head = list->tail = list->cur = NULL;
  list->n_items = 0;
  return list;
}

real *RealListToVector
  (struct REAL_LIST *list,
   int *n)
{
  int i = 0;
  real *v = CallocOrAbort(real, list->n_items);
  struct REAL_ITEM *p;
  for (p=list->head; p!=NULL; p=p->next)
    v[i++] = p->x;
  *n = list->n_items;
  return v;
}


/***
 * Skip ``small white space'' in a file, where small white space is
 * space, a tab, or a # and anything following it * on the same line.
 * Returns 1 if neither EOF or EOLN is reached,
 *         0 if EOLN is reached,
 *         -1 if EOF is reached.
 */

int SkipLineSpace
  (FILE *file)
{
  int c;
  while (1) {
    c = getc(file);
    if (c==EOF)
      return EOF;
    else if (c==' ' || c=='\t')
      continue;
    else if (c=='#') {
      int lastc;
      c = 0;
      /* read to the end of the line */
      do {
	lastc = c;
	c = getc(file);
	if (c==EOF)
	  return EOF;
      } while (c!='\n');
      /* return if the EOLN was not preceeded by a backslash */
      if (lastc!='\\')
	return 1;
    } else if (c=='\\') {
      c = getc(file);
      if (c==EOF)
	return EOF;
      if (!isspace(c)) {
	ungetc(c, file);
	if (ungetc('\\', file)==EOF)
	  ErrorAbort("SkipSmallSpace: cannot cope with \\%c in input", c);
	return 1;
      }
    } else {
      ungetc(c, file);
      if (c=='\n')
	return 0;
      else
	return 1;
    }
  }
}

/***
 * Skip ``white space'' in a file, where white space is
 * space, tab, newline, or a # and anything following it
 * on the same line.
 * Returns 1 if EOF is not reached, and EOF if EOF is reached
 */

int SkipSpace
  (FILE *file)
{
  int c;
  while (1) {
    c = getc(file);
    if (c==EOF)
      return EOF;
    else if (isspace(c))
      continue;
    else if (c=='#') {
      /* read to the end of the line */
      do {
	c = getc(file);
	if (c==EOF)
	  return EOF;
      } while (c!='\n');
    } else {
      ungetc(c, file);
      return 1;
    }
  }
}

/***
 * Reads a vector from the given file.
 * Tries to be a bit smart about the format, and space allocation,
 * depending on the parameters passed.
 *
 * Alternatives:
 *   *np==0 && v==NULL
 *     Reads a vector from one line of the file, allocates storage for it,
 *     and sets *np to the length of the vector.
 *   *np==0 && v!=NULL
 *     Error - this doesn't make sense.
 *   *np==n && v==NULL
 *     Reads a vector of length n from the file, and allocates storage for it.
 *   *np==n && v!=NULL
 *     Reads a vector of length n from the file into v
 * If EOF is reached before any elements are read, NULL is returned.
 * In other cases, an unexpected EOF causes an ErrorAbort.
 */

real *ReadVector
  (FILE *file,
   int *np,
   real *v)
{
  int i;
  double x;
  extern int errno;
  if (*np>0) {
    /***
     * We are given the length of the vector
     */
    if (v==NULL)
      v = CallocOrAbort(real, *np);
    for (i=0; i<*np; i++) {
      if (SkipSpace(file)!=1) {
        if (feof(file)) {
	  if (i==0)
	    return NULL;
	  else
	    ErrorAbort("ReadVector: unexpected EOF reading element %d", i);
	} else {
	  ErrorAbort("ReadVector: %s", strerror(errno));
        }
      }
      if (fscanf(file, "%lf", &x)!=1)
	ErrorAbort("ReadVector: bad format reading element %d", i);
      v[i] = x;
    }
  } else {
    /***
     * Have to work out the length of the vector.
     */
    struct REAL_LIST *list;
    if (SkipSpace(file)!=1) {
      if (feof(file))
	return NULL;
      else
	ErrorAbort("ReadVector: %s", strerror(errno));
    }
    list = NewRealList();
    while (1) {
      /* skip spaces and tabs, and newlines preceeded by backslashes */
      if (SkipLineSpace(file)!=1)
	break;
      if (fscanf(file, "%lf", &x)!=1)
	break;
      else
	AddRealItem(list, (real) x);
    }
    if (list->n_items==0)
      return NULL;
    else {
      v = RealListToVector(list, np);
      FreeRealList(list);
    }
  }
  return v;
}

/***
 * Reads a label from the given file.  A label must start with a non-digit.
 * Mallocs the space, in 8 byte increments
 */

#define INCREMENT 8

char *ReadLabel
  (FILE *file)
{
  int c, str_len, alloc_len, need_len;
  char *str = NULL;
  if (SkipSpace(file)!=1) {
    if (feof(file))
      return (char*)EOF;
    else
      ErrorAbort("ReadLabel: %s", strerror(errno));
  }
  c = getc(file);
  if (c==EOF)
    return (char*)EOF;
  if (isdigit(c) || (c=='.') || (c=='-')) {
    ungetc(c, file);
    return NULL;
  }
  str_len = 1;
  alloc_len = 0;
  do {
    /* alloc_len has to be greater than str_len+1, and must be a multiple
       of INCREMENT */
    need_len = (((str_len+1+INCREMENT-1) / INCREMENT) * INCREMENT);
    /* printf("%d %d %d %s\n", str_len, alloc_len, need_len, str); */
    if (need_len != alloc_len) {
      if (str==NULL)
	str = CallocOrAbort(char, need_len);
      else
	str = ReallocOrAbort(char, str, need_len);
      alloc_len = need_len;
    }
    str[str_len-1] = c;
    str[str_len] = '\0';
    str_len++;
    c = getc(file);
  } while (c!=EOF && !isspace(c));
  return str;
}

/***
 * NextVector(file, x_np, xp, labelp, vnump)
 *
 * If x is a vector, and vnum is 0, return with 1 (this is to
 *   handle a command line vector.
 * Otherwise, try to read a vector from the file.
 *   If successful, return with 1
 *   Otherwise, return with 0, or abort if there is a read error.
 *
 * If dimensionality is supplied (in x_n), use it, otherwise
 * read a vector from one line.
 */

int NextVector
  (FILE *file,
   int *x_np,
   real **xp,
   char **labelp,
   int *vnump)
{
  if (*xp==NULL || vnump==NULL || *vnump>0) {
    /* want to read a vector from the file */
    if (file==NULL)
      return 0;
    if (SkipSpace(file)!=1)
      if (feof(file))
	return 0;
      else
	ErrorAbort("NextVector: %s", strerror(errno));
    else {
      if (labelp!=NULL)
	*labelp = ReadLabel(file);
      if (feof(file))
	ErrorAbort("EOF reached while reading vector");
      *xp = ReadVector(file, x_np, *xp);
    }
  }
  if (*x_np==0 || *xp==NULL) {
    if (ferror(file))
      ErrorAbort("NextVector: %s", strerror(errno));
    else if (!feof(file)) {
      fprintf(stderr, "Warning: EOF not reached and vector not read.\n");
      return 0;
    } else
      return 0;
  }

  if (vnump!=NULL)
    (*vnump)++;
  return 1;
}

/***
 * real **readMatrix(file, rowsp, colsp, M, labelvp)
 */
real **readMatrix
  (FILE *file,
   int *rowsp,
   int *colsp,
   real **M,
   char ***labelvp)
{
  int i, vnum = 0;
  real *v = NULL;
  char *label;
  int somelabel = 0;
  if (*rowsp>0) {
    if (M==NULL)
      M = CallocOrAbort(real *, *rowsp);
    if (*labelvp==NULL)
      *labelvp = CallocOrAbort(char *, *rowsp);
    for (i=0; i<*rowsp; i++)
      if (!NextVector(file, colsp, &(M[i]),
		      labelvp==NULL ? NULL : &((*labelvp)[i]), &vnum))
	ErrorAbort("readMatrix: cannot read row %d\n", i);
    return M;
  } else {
    struct POINTER_LIST *vlist, *llist=NULL;

    if (labelvp!=NULL)
      llist = NewPointerList();
    vlist = NewPointerList();

    while (NextVector(file, colsp, &v, labelvp==NULL ? NULL : &label, &vnum)) {
      AddPointerItem(vlist, (char*) v);
      if (labelvp!=NULL) {
	AddPointerItem(llist, label);
	if (label!=NULL && label[0]!='\0')
	  somelabel = 1;
      }
      v = NULL;
    }

    M = (real **) PointerListToVector(vlist, rowsp);
    FreePointerList(vlist);
    if (labelvp!=NULL) {
      if (somelabel)
	*labelvp = PointerListToVector(llist, NULL);
      else
	*labelvp = NULL;
      FreePointerList(llist);
    }

    return M;
  }
}

/***
 * Call with vectors indexed [0..n-1]
 */
real *convolve
  (int n,
   real *z,
   real *x,
   real *y)
{
  int i, j, k;
  for (k=0;k<n;k++) {
    z[k] = 0.0;
    j = k;
    /* fprintf(dout, "convolve   z[%d] =", k); */
    for (i=0;i<n;i++) {
      /* fprintf(dout, " x[%d]*y[%d]", i, j, x[i], y[j]); */
      z[k] += x[i]*y[j];
      j--;
      if (j<0) j = n-1;
    }
    /* fprintf(dout, "\n"); */
  }
  return z;
}

/***
 * Does a circular convolution of data1 and data2 via FFTs
 *
 * Call with vectors indexed [0..n-1]
 */
real *convolveFFT
  (int n,
   real ans[],
   real data1[],
   real data2[])
{
  int i;
  static real *fft1=NULL, *fft2=NULL;
  static int last_n = 0;

  if (last_n==0 || last_n!=n) {
    if (last_n!=0) {
      free(fft1);
      free(fft2);
    }
    fft1 = CallocOrAbort(real, n);
    fft2 = CallocOrAbort(real, n);
    last_n = n;
  }

  for (i=0; i<n; i++) {
    fft1[i] = data1[i];
    fft2[i] = data2[i];
  }
  rfft(fft1, n);
  rfft(fft2, n);

  ans[0] = (fft1[0] * fft2[0]);
  ans[1] = (fft1[1] * fft2[1]);
  for (i=2;i<n;i+=2) {
    ans[i] = (fft1[i]*fft2[i] - fft1[i+1]*fft2[i+1]);
    ans[i+1] = (fft1[i]*fft2[i+1] + fft1[i+1]*fft2[i]);
  }

  rifft(ans, n);

  return ans;
}

#ifdef OBSELETE
/* Old version that used NumRec fft's */
real * ConvolveFFT(n, ans, data1, data2)
real data1[], data2[], ans[];
int n;
{
  int i;
  static real *fft1, *fft2;
  static int last_n = 0;
  void twofft(), realft(), nrerror();
  data1--;
  data2--;
  ans--;

  if (last_n==0 || last_n!=n) {
    if (last_n!=0) {
      FreeVector(real, fft1, 1, last_n);
      FreeVector(real, fft2, 1, last_n);
    }
    fft1 = Vector(real, 1, n);
    fft2 = Vector(real, 1, n);
    last_n = n;
  }

  fft2re(data1, data2, fft1, fft2, n);

  ans[1] = 2*(fft1[1] * fft2[1])/n;
  ans[2] = 2*(fft1[2] * fft2[2])/n;
  for (i=3;i<n;i+=2) {
    ans[i] = 2*(fft1[i]*fft2[i] - fft1[i+1]*fft2[i+1])/n;
    ans[i+1] = 2*(fft1[i]*fft2[i+1] + fft1[i+1]*fft2[i])/n;
  }

  realft(ans, n/2, -1);

  return ans+1;
}
#endif

/***
 * PowerConv - compute y = x to the convolative power t,
 * x is in the time domain.
 */
real *convolvePow(int n, real *y, real *x, real t)
{
  int i;
  real r, theta, new_r, new_theta;
  /* transform into freq domain */
  for (i=0; i<n; i++)
    y[i] = x[i];
  rfft(y, n);
  /* compute power - in freq. domain take inverse of each element */
  y[0] = pow(y[0], t);
  y[1] = pow(y[1], t);
  for (i=2 ; i<n; i+=2) {
    /* get polar representation */
    r		= hypot(y[i], y[i+1]);
    theta	= (r==0) ? 0 : acos(y[i]/r) * (y[i+1]>0.0 ? 1.0 : -1.0);
    /* compute power */
    new_r	= pow(r, t);
    new_theta	= fmod(t*theta, 2.0*M_PI);
    /* and back into rectangular */
    y[i]	= new_r * cos(new_theta);
    y[i+1]	= new_r * sin(new_theta);
  }
  rifft(y, n);
  return y;
}

/***
 * Call with vectors indexed [0..n-1]
 */
real *correlate
  (int n,
   real *z,
   real *x,
   real *y)
{
  int i, j, k;

  for (k=0;k<n;k++) {
    z[k] = 0.0;
    j = k;
    /* fprintf(dout, "correlate z[%d] =", k); */
    for (i=0;i<n;i++) {
      /* fprintf(dout, " x[%d]*y[%d]", i, j, x[i], y[j]); */
      z[k] += x[i]*y[j];
      j++;
      if (j>=n) j = 0;
    }
    /* fprintf(dout, "\n"); */
  }
  return z;
}

/***
 * Call with vectors indexed [0..n-1]
 *
 * Convolves data2 with the involution of data1
 */
real *correlateFFT
  (int n,
   real ans[],
   real data1[],
   real data2[])
{
  int i;
  static real *fft1, *fft2;
  static int last_n = 0;

  if (last_n==0 || last_n!=n) {
    if (last_n!=0) {
      free(fft1);
      free(fft2);
    }
    fft1 = CallocOrAbort(real, n);
    fft2 = CallocOrAbort(real, n);
    last_n = n;
  }

  for (i=0; i<n; i++) {
    fft1[i] = data1[i];
    fft2[i] = data2[i];
  }
  rfft(fft1, n);
  rfft(fft2, n);

  ans[0] = (fft1[0] * fft2[0]);
  ans[1] = (fft1[1] * fft2[1]);
  for (i=2;i<n;i+=2) {
    ans[i] = (fft1[i]*fft2[i] + fft1[i+1]*fft2[i+1]);
    ans[i+1] = (fft1[i]*fft2[i+1] - fft1[i+1]*fft2[i]);
  }

  rifft(ans, n);

  return ans;
}

#ifdef OBSELETE
/* Old version that uses NumRec fft's */

real * CorrelateFFT(n, ans, data1, data2)
real data1[], data2[], ans[];
int n;
{
  int i;
  static real *fft1, *fft2;
  static int last_n = 0;
  void twofft(), realft(), nrerror();
#ifdef DBMALLOC
  malloc_check_data("ConvolveFFT", ans, n*sizeof(real));
  malloc_check_data("ConvolveFFT", data1, n*sizeof(real));
  malloc_check_data("ConvolveFFT", data2, n*sizeof(real));
#endif /* DBMALLOC */
  data1--;
  data2--;
  ans--;

  if (last_n==0 || last_n!=n) {
    if (last_n!=0) {
      FreeVector(real, fft1, 1, last_n);
      FreeVector(real, fft2, 1, last_n);
    }
    fft1 = Vector(real, 1, n);
    fft2 = Vector(real, 1, n);
    last_n = n;
  }

  fft2re(data1, data2, fft1, fft2, n);

  ans[1] = 2*(fft1[1] * fft2[1])/n;
  ans[2] = 2*(fft1[2] * fft2[2])/n;
  for (i=3;i<n;i+=2) {
    ans[i] = 2*(fft1[i]*fft2[i] + fft1[i+1]*fft2[i+1])/n;
    ans[i+1] = 2*(fft1[i]*fft2[i+1] - fft1[i+1]*fft2[i])/n;
  }

  realft(ans, n/2, -1);

  return ans+1;
}
#endif

real dotProduct
  (int n,
   real *x,
   real *y)
{
  int i;
  double z = 0.0;
  if (x!=y)
    for (i=0; i<n; i++)
      z += x[i] * y[i];
  else
    for (i=0; i<n; i++)
      z += x[i] * x[i];
  return z;
}

real cosine
  (int n,
   real *x,
   real *y)
{
  real x_size = vSize(n, x);
  real y_size = vSize(n, y);
  double xy_sum = 0;
  int i;
  for (i=0; i<n; i++)
    xy_sum += x[i]*y[i];
  return xy_sum/(x_size*y_size);
}

/***
 * Returns the Euclidean length of the vector
 * v[0..n-1]
 * Avoid overflow by using a generalization of the technique
 * sqrt(x^2 + y^2) = x * sqrt(1 + (y/x)^2), where fabs(x)>fabs(y)
 */
real vSize(int n, real *v)
{
  int i;
  double sum = 1.0;
  double multiplier = fabs(v[0]);
  for (i=1;i<n;i++) {
    if (fabs(v[i]) > multiplier) {
      /* rescale the sum: sum' = sum*multiplier^2/v[i]^2 */
      sum = sum*(multiplier/v[i])*(multiplier/v[i]) + 1.0;
      multiplier = fabs(v[i]);
    } else {
      if (multiplier!=0.0)
	sum += (v[i]/multiplier)*(v[i]/multiplier);
    }
  }
  return sqrt(sum)*multiplier;
}

