#include <stdlib.h>
#include <string.h>
#include "utils.h"
#include "real.h"
#include "hashTable.h"
#include "config.h"
#ifndef NULL
#define NULL (0L)
#endif

/*
 *  HashTable.c -	implements a general hash table
 *
 *  items in the table have four fields: name, context, id and value
 *  the context and id fields are integers, the name is a string, and
 *  the value field is an array of 8 integers
 *  items can be retrieved on any of the three first fields
 *  name and context together form a unique key, as does id by itself
 */

/*
 * TODO: make insert look for id's of deleted items when it fills up
 * maybe keep a count of live items to assist this
 */

#define MAX_STRING (1024)
int hash_errno;
char *hash_error[10] = {
  "everything is ok, there's no error!",
  "name not found in table",
  "name is ambiguous",
  "name is already in table",
  "id out of range",
  "out of memory",
  "name is null or zero length",
  "hash table overflow",
  "not configured for performing this task",
  NULL};

int hash(char *str,
	 int max_hash,
	 int context)
{
  int h = 0;
  int n;
  while ((n = *(str++))!=0) {
    h = h*11 + (n-'a');
  }
  h += context;
  h %= max_hash;
  if (h<0) h = -h;
  return h;
}

HashTable *HashTableNewTable
  (HashTable *ht,
   int max_items,	/* int - maximum number of items */
   int hash_bins,	/* int - size of the hash table */
   int index_types)	/* can be ID_INDEX or ALPHA_INDEX */
{
  hash_errno = 0;
  if (hash_bins<1) hash_bins=1;
  if (ht==NULL)
    if((ht = (HashTable*)malloc(sizeof(HashTable)))==NULL)
      goto insuff_mem;
  ht->id_vec = NULL;
  ht->alpha_vec = NULL;
  ht->hash_bins = hash_bins;
  if ((ht->hash_vec=Calloc(HashTableItem*, hash_bins))==NULL)
    goto insuff_mem;
  if ((index_types&ID_INDEX)
      && ((ht->id_vec=Calloc(HashTableItem*, max_items))==NULL))
    goto insuff_mem;
  ht->max_items = max_items;
  ht->free_id = 1;
  ht->free_context = 1;
  ht->free_item = NULL;
  return ht;

 insuff_mem:
  ErrorAbort("HashTableNewTable: out of memory");
  hash_errno = HT_NO_MEMORY;
  return NULL;
}

HashTableItem *HashTableLookup
  (HashTable *ht,
   char *name,
   int context)
{
  int hv;
  HashTableItem *item, *prev_item;
  hash_errno = 0;
  if (name==NULL || name[0]==0) {
    hash_errno = HT_BAD_NAME;
    return NULL;
  }
  /*
   * look with the hash table
   */
  hv = hash(name, ht->hash_bins, context);
  item = ht->hash_vec[hv];
  prev_item = 0;
  while (item!=NULL) {
    if ((context==item->context) && !strcmp(name, item->name)) {
      /* put the found item at the start of the list */
      if (prev_item!=0) {
	prev_item->next = prev_item->next->next;
	item->next = ht->hash_vec[hv];
	ht->hash_vec[hv] = item;
      }
      return item;
    }
    prev_item = item;
    item = item->next;
  }
  hash_errno = HT_NOT_FOUND;
  return NULL;
}

#ifdef OBSELETE
static int ht_lookup_id(ht, name, context)
HashTable *ht;
char *name;
int context;
{
  int hv;
  HashTableItem *item, *prev_item;
  hash_errno = 0;
  if (name==NULL || name[0]==0) {
    hash_errno = HT_BAD_NAME;
    return NULL;
  }
  /*
   * look with the hash table
   */
  hv = hash(name, ht->hash_bins, context);
  item = ht->hash_vec[hv];
  prev_item = 0;
  while (item!=NULL) {
    if ((context==item->context) && !strcmp(name, item->name)) {
      /* put the found item at the start of the list */
      if (prev_item!=0) {
	prev_item->next = prev_item->next->next;
	item->next = ht->hash_vec[hv];
	ht->hash_vec[hv] = item;
      }
      return item->id;
    }
    prev_item = item;
    item = item->next;
  }
  hash_errno = HT_NOT_FOUND;
  return 0;
}
#endif

int HashTableNewContext(HashTable *ht)
{
  return ht->free_context++;
}

HashTableItem *HashTableLookupById(HashTable *ht, int id)
{
  int i;
  HashTableItem *item;
  hash_errno = 0;
  if (id<1 || id>=ht->max_items) {
    hash_errno = HT_BAD_ID;
    return NULL;
  }
  /*
   * use the id vector to return it instantly
   */
  if (ht->id_vec!=NULL) return ht->id_vec[id];
  /*
   * or, if that doesn't exits, laboriously traverse
   * the hash table
   */
  i = 0;
  while (i<ht->hash_bins) {
    item = ht->hash_vec[i];
    while (item!=NULL) if (item->id==id) return item; else item=item->next;
    i++;
  }
  ErrorAbort("HashTableLookupById : couldn't find id %d", id);
  hash_errno = HT_NOT_FOUND;
  return NULL;
}

#define EXTEND 16

HashTableItem *HashTableInsert
  (HashTable *ht,
   char *name,
   int context,
   void *data)
{
  int len, hv, found;
  HashTableItem *item;

  hash_errno = 0;
  /* check if the string is unreasonable small or large */
  if (((len=strlen(name))==0) || len >= MAX_STRING) {
    hash_errno = HT_BAD_NAME;
    return NULL;
  }

  if (ht->free_id>=ht->max_items && ht->id_vec!=NULL) {
    /* see if we can find any free ids */
    ht->free_id = 1;
    while (ht->id_vec[ht->free_id]!=NULL && ht->free_id<ht->max_items) {
      /* fprintf(stderr, "id %d not free\n", ht->free_id); */
      ht->free_id++;
    }
  }
  if (ht->free_id>=ht->max_items) {
    /* reallocate hash, id, & alpha vector instead of returning NULL */
    HashTableItem **vec;
    int i;
    if (ht->id_vec!=NULL) {
      vec = Realloc(HashTableItem*, ht->id_vec, ht->max_items+EXTEND);
      if (vec==NULL) goto out_of_mem;
      for (i=0;i<EXTEND;i++) vec[ht->max_items+i] = NULL;
      ht->id_vec = vec;
    }
    if (ht->alpha_vec!=NULL) {
      vec = Realloc(HashTableItem*, ht->alpha_vec, ht->max_items+EXTEND);
      if (vec==NULL) goto out_of_mem;
      for (i=0;i<EXTEND;i++) vec[ht->max_items+i] = NULL;
      ht->alpha_vec = vec;
    }
    ht->max_items += EXTEND;
    goto OK;

  out_of_mem:
    hash_errno = HT_OVERFLOW;
    return NULL;
  OK:;
  }

  hv = hash(name, ht->hash_bins, context);
  item = ht->hash_vec[hv];
  found=0;
  while (item!=NULL && !found) {
    if ((context==item->context) && !strcmp(name, item->name))
      found=1;
    else
      item = item->next;
  }
  if (found) {
    hash_errno = HT_DUPLICATE;
    return NULL;
  }

  item = Calloc(HashTableItem, 1);
  if (item==NULL) {
    hash_errno = HT_NO_MEMORY;
    return NULL;
  }
  item->name=strdup(name);
  if (item->name==NULL) {
    /* can't be bothered to delete the item, leave it around as trash */
    hash_errno = HT_NO_MEMORY;
    return NULL;
  }
  item->id = ht->free_id;
  ht->free_id++;
  if (ht->id_vec!=NULL)
    while (ht->id_vec[ht->free_id]!=NULL && ht->free_id<ht->max_items)
      ht->free_id++;
  /* fprintf(stderr, "new hash table entry, id %d\n", item->id); */
  item->context = context;
  item->data.ptr = data;
  if (ht->id_vec!=NULL)
    ht->id_vec[item->id] = item;
  item->next = ht->hash_vec[hv];
  ht->hash_vec[hv] = item;
  return item;
}

int HashTableDelete
  (HashTable *ht,
   char *name,
   int context)
{
  int hv, found;
  HashTableItem *item, *p;

  hv = hash(name, ht->hash_bins, context);
  item = ht->hash_vec[hv];
  found=0;
  p=NULL;
  while (item!=NULL && !found) {
    if ((context==item->context) && !strcmp(name, item->name))
      found=1;
    else {
      p=item;
      item = item->next;
    }
  }
  if (!found) return 0;
  else {
    if (p==NULL)
      ht->hash_vec[hv] = ht->hash_vec[hv]->next;
    else
      p->next = p->next->next;
    if (ht->id_vec!=NULL)
      ht->id_vec[item->id] = NULL;
    /* fprintf(stderr, "deleted item %d\n", item->id); */
    /* modify alpha_vec here */
    if (ht->free_item!=NULL)
      (ht->free_item)(item->data.ptr);
    free(item->name);
    free((char*)item);
  }
  return 1;
}

void HashTablePrint(ht)
HashTable *ht;
{
  int i;
  HashTableItem *item;
  printf("Hash table, %d locations\n", ht->hash_bins);
  for (i=0;i<ht->hash_bins;i++) {
    item = ht->hash_vec[i];
    printf("VEC[%d] ", i);
    while (item!=NULL) {
      printf("(%s, %d, %d) ", item->name, item->context, item->id);
      item=item->next;
    }
    printf("\n");
  }
}

HashTableItem *HashTableNext(HashTable *ht, int *id)
{
  if (ht->id_vec == 0) {
    ErrorAbort("HashTableNext : no id vector, cannot do");
    return 0;
  }
  while (++(*id) < ht->max_items) /* skip holes (made by delete) */
    if (ht->id_vec[(*id)]!=NULL) return ht->id_vec[(*id)];
  *id = 0;
  return 0;
}

HashTableItem *HashTableNextByContext (HashTable *ht, int *id, int context)
{
  hash_errno = 0;
  if (ht->id_vec == 0) {
    ErrorAbort("HashTableNextByContext : no id vector, cannot do");
    return NULL;
  }
  while (++(*id) < ht->max_items) /* skip holes (made by delete) */
    if (   ht->id_vec[(*id)]!=NULL
	&& (context==0 || ht->id_vec[(*id)]->context==context))
      return ht->id_vec[(*id)];
  *id = 0;
  return NULL;
}

int HashTableDeleteByContext(HashTable *ht, int context)
{
  int i, c;
  HashTableItem *prev, *cur, *next;
  c = 0;
  hash_errno = 0;
  for (i=0;i<ht->hash_bins;i++) {
    cur = prev = ht->hash_vec[i];
    while (cur!=NULL) {
      next = cur->next;
      if (cur->context == context) {
	c++;
	/* adjust the pointers */
	if (cur==prev)
	  ht->hash_vec[i] = cur->next;
	else
	  prev->next = cur->next;
	/* adjust the storage vectors */
	if (ht->id_vec!=NULL)
	  ht->id_vec[cur->id] = NULL;
	/* fprintf(stderr, "deleted item %d\n", cur->id); */
	/* modify alpha vec here */
	/* free the item */
	if (ht->free_item!=NULL)
	  (ht->free_item)((char*)cur->data.ptr);
	free(cur->name);
	free((char*)cur);
      } else {
	prev = cur;
      }
      cur = next;
    }
  }
  return c;
}
