#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <string.h>
#include <math.h>
#include <values.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#ifdef USE_TAP_LIBRARIES
  #define NO_String_TYPE
  #include <tap/argf.h>
  #include <tap/tapmath.h>
  #include <tap/taputil.h>
  #include <tap/HashTable.h>
  #include <tap/fft.h>
#else
  #include "config.h"
  #include "argf.h"
  #include "utils.h"
  #include "real.h"
  #include "hashTable.h"
  #include "fft.h"
  #include "functions.h"
#endif /* USE_TAP_LIBRARIES */
#include "vcalc.h"
#include "calcgram.tab.h"

int param_context;	/* context for parameters in hash table */
int var_context;	/* context for var in hash table */
int local_context;	/* context for local vars in hash table */
int debugAlloc = 0;
struct HashTable *ht;
extern FILE *yyin;
extern int yylex();
extern char yytext[];
extern struct ExprNode *ParseExpr(void);
jmp_buf abort_expr_env;
char *inputLine;
char *inputLineEnd;
void **freeList;
int freeListTop = 0;
int freeListUpper = 0;
#define FREE_LIST_EXTEND 16
enum NORM_TYPE {LOGISTIC, HEAVISIDE, UNITNORM, UNITBINARY, POWERNORM, TANH, SYMMETRIC, NOPNORM, LINEAR, BOUNDEDLINEAR};
enum SIM_TYPE {DOTPRODUCT, COSINE};
enum RAND_DIST {UNIFORM, NORMAL};

/* Stuff for the bit counting table */
/* Works faster counting 8 bit chunks rather than 16 or 4 */
/* Probably because it can keep the 8 bit table in cache */
#define HAVE_BIT_CHUNK_TYPE /* this one is optional */
typedef unsigned char bitChunk;
#define BIT_COUNT_CHUNK_SIZE 8
#define BIT_COUNT_TABLE_SIZE (1<<BIT_COUNT_CHUNK_SIZE)
int bitCountTable[BIT_COUNT_TABLE_SIZE];
#define BIT_COUNT_CHUNK_MASK ((1 << BIT_COUNT_CHUNK_SIZE) - 1)
#define BIT_CHUNKS_PER_WORD (((BITS(bitWord)-1)/BIT_COUNT_CHUNK_SIZE)+1)

/* forward declarations */
struct VarNode *func_store(struct ExprNode *node);
struct VarNode *func_match(struct ExprNode *node);
struct VarNode *func_acam(struct ExprNode *node, char *what);
struct VarNode *func_blend(struct ExprNode *node);
struct VarNode *func_orthogonalize(struct ExprNode *node);
struct VarNode *func_push(struct ExprNode *node);
struct VarNode *func_pop(struct ExprNode *node);
struct VarNode *func_top(struct ExprNode *node);
struct VarNode *func_cadr(char *what, struct ExprNode *node);
struct VarNode *func_cons(struct ExprNode *node);
struct VarNode *func_cosine(struct ExprNode *node);
struct VarNode *func_sqrt(struct ExprNode *node);
struct VarNode *func_time(struct ExprNode *node);
struct VarNode *func_rand(struct ExprNode *node);
struct VarNode *func_uniform(struct ExprNode *node);
struct VarNode *func_randVec(struct ExprNode *node);
struct VarNode *func_bitvec(struct ExprNode *node);
struct VarNode *func_randSeed(struct ExprNode *node);
struct VarNode *func_repeat(struct ExprNode *node);
struct VarNode *func_print(struct ExprNode *node);
struct VarNode *func_normalize(struct ExprNode *node);
struct VarNode *func_varAlias(struct ExprNode *node);
struct VarNode *makePermanent(struct VarNode *var);
struct VarNode *makeConstant(struct VarNode *var);
struct VarNode *lookupVar(char *name);
struct VarNode *newScalarVar(real x);
struct VarNode *newStringVar(char *str);
struct VarNode *newVectorVar(int n, real *v);
struct VarNode *newBitVectorVar(int n, bitWord *b);
struct VarNode *newLNodeVar(struct VarNode *car, struct VarNode *cdr);
void freeNode(struct ExprNode *node);
void freeVar(struct VarNode *var);
void freeLocalVars(void);
void addToFreeList(void *ptr);
void freeFreeList(void);
void printTree(struct ExprNode *node);
int suppressPrinting(struct ExprNode *node);
void printDebugTree(struct ExprNode *node, int depth);
void evalTree(struct ExprNode *node);
void evalFunction(struct ExprNode *node, char *name);
void doAssignment(struct ExprNode *node);
void printOp(int op);
void printVarValue(struct VarNode *var, int level);
void normalizeVec(int n, real *x, real *result);
int cadrString(char *str);
int trueString(char *str);
real scalarVarValue(char *str);
char * stringVarValue(char *str);
enum NORM_TYPE normType(char *);
enum SIM_TYPE simType(char *);
int standardVectorType(int);
void initBitCountTable(void);
int eof_flag;
FILE **files = NULL;
int nfile = 0;

int main(int argc, char *argv[])
{
  int rows = 0, cols = 0;
  int cosine = 0;
  real **M = NULL;
  char **matlabels = NULL;
  int i;
  char *format = "%.3g";
  char *labelformat = "%-1s ";
  char *separator = " ";
  int horiz = 2;
  /* volatile here to that GCC doesn't potentially clobber this
   * variable on longjmp, or more importantly give annoying warnings
   about it */
  volatile int done = 0;
  int seed = 0;
  int echo = 0;
  int storeBitVectors = 1;
  int debugTree = 0;
  int debugAddr = 0;
  int debug = 0;
  volatile int debugArenaWarned = 0;
  volatile int usePrompt = 0;
  volatile int suppressPrompt = 0;
  int dim = 8;
  int circular = 0;
  FILE *matfile = NULL;
  struct ExprNode *expr;
  struct VarNode *var;
  struct ARG_DESC *argd;
  dout = stdout;
  din = stdin;

#ifdef ARENA_MEM
  initAlloc();
#endif

  sizeOfReal = sizeof(Real);
  argd = StartArgs(argv[0]);
  Args(argd, "[-Pvectors <vector file>%I%?]", &matfile,
       "file to read vectors from");
  Args(argd, "[<file>...%I%?]", &files, "files to read commands from");
  Args(argd, "[-Pdim <n>%i%?]", &dim, "dimensionality of vectors in file");
  Args(argd, "[-Pseed <n>%i%?]", &seed, "seed for random number generator");
  Args(argd, "[-Phoriz <n>%i%?]", &horiz, "width for each horizontal label");
  Args(argd, "[-Pcosine %P%?]", &cosine,
       "use cosine rather than dot-prod to compare");
  Args(argd, "[-Pcircular %P%?]", &circular,
       "use the circular vector model");
  Args(argd, "[-Sdebug %S%?]", &debug, "general debugging");
  Args(argd, "[-Sdebugtree %S%?]", &debugTree, "print parse trees");
  Args(argd, "[-Secho %S%?]", &echo, "echo input");
  Args(argd, "[-SstoreBitVectors %S%?]", &storeBitVectors, "use bit vectors in stores");
  Args(argd, "[-SdebugAlloc %S%?]", &debugAlloc, "log all calls to alloc and free");
  Args(argd, "[-SdebugAddr %S%?]", &debugAddr, "print trees with memory addresses");
  Args(argd, "[<format>%F%?]", &format,
       "format to print vector elements, default \"%.6g\"");
  Args(argd, "[-Plabelformat <format>%F%?]", &labelformat,
       "format to print labels, default \"%-1s \"");
  Args(argd, "[-Pseparator <string>%s%?]", &separator,
       "string to separate vector elements, default \" \"");
  EndArgs(argd);
  argv = ParseArgs(argd, argc, argv, 0);

  inputLine = CallocOrAbort(char, 8096);
  srand((int) seed);
  srandom(rand());

  rows = 0;
  if (matfile!=NULL) {
    M = readMatrix(matfile, &rows, &cols, M, &matlabels);
    if (rows==0)
      ErrorAbort("no vectors read from file");
    if (matlabels==NULL)
      ErrorAbort("vectors must have labels");
  }

  ht = HashTableNewTable(NULL, 40, 500, ID_INDEX);
  HashTableDeAllocator(ht, (HashTableDataDeAllocater) freeVar);
  param_context = HashTableNewContext(ht);
  var_context = param_context+1;
  local_context = param_context+2;

  HashTableInsert(ht, "echo", param_context,
		  makePermanent(newStringVar(echo ? "true" : "false")));
  HashTableInsert(ht, "store.bitvectors", param_context,
		  makePermanent(newStringVar(storeBitVectors ? "true" : "false")));
  HashTableInsert(ht, "match.sim", param_context,
		  makePermanent(newStringVar("dotProduct")));
  HashTableInsert(ht, "match.print", param_context,
		  makePermanent(newStringVar("true")));
  HashTableInsert(ht, "match.filter", param_context,
		  makePermanent(newStringVar("false")));
  HashTableInsert(ht, "match.spread", param_context,
		  makePermanent(newScalarVar(4.0)));
  HashTableInsert(ht, "match.precision", param_context,
		  makePermanent(newScalarVar(3.0)));
  HashTableInsert(ht, "match.capacity", param_context,
		  makePermanent(newScalarVar(1.0)));
  HashTableInsert(ht, "blend.sim", param_context,
		  makePermanent(newStringVar("dotProduct")));
  HashTableInsert(ht, "blend.transfer", param_context,
		  makePermanent(newStringVar("sigmoid")));
  HashTableInsert(ht, "blend.bias", param_context,
		  makePermanent(newScalarVar(0.0)));
  HashTableInsert(ht, "blend.threshhold", param_context,
		  makePermanent(newScalarVar(0.0)));
  HashTableInsert(ht, "blend.weight", param_context,
		  makePermanent(newScalarVar(1.0)));
  HashTableInsert(ht, "blend.capacity", param_context,
		  makePermanent(newScalarVar(1.0)));
  HashTableInsert(ht, "retrieve.sim", param_context,
		  makePermanent(newStringVar("dotProduct")));
  HashTableInsert(ht, "retrieve.transfer", param_context,
		  makePermanent(newStringVar("binary")));
  HashTableInsert(ht, "retrieve.bias", param_context,
		  makePermanent(newScalarVar(0.0)));
  HashTableInsert(ht, "retrieve.threshhold", param_context,
		  makePermanent(newScalarVar(0.0)));
  HashTableInsert(ht, "retrieve.weight", param_context,
		  makePermanent(newScalarVar(1.0)));
  HashTableInsert(ht, "retrieve.capacity", param_context,
		  makePermanent(newScalarVar(1.0)));
  HashTableInsert(ht, "rand.seed", param_context,
		  makePermanent(newScalarVar((real) seed)));
  var = makePermanent(newScalarVar((real) dim));
  HashTableInsert(ht, "var.dim", param_context, var);
  HashTableInsert(ht, "vector.dim", param_context, var);
  HashTableInsert(ht, "randvec.dim", param_context, var);
  HashTableInsert(ht, "randvec.normalize", param_context,
		  makePermanent(newStringVar("false")));
  HashTableInsert(ht, "normalize.type", param_context,
		  makePermanent(newStringVar("unit")));
  HashTableInsert(ht, "normalize.threshhold", param_context,
		  makePermanent(newScalarVar(0.0)));
  HashTableInsert(ht, "vector.type", param_context,
		  makePermanent(newStringVar(circular?"circular":"standard")));
  HashTableInsert(ht, "conv.fft", param_context,
		  makePermanent(newStringVar("true")));
  HashTableInsert(ht, "print.brace", param_context,
		  makePermanent(newStringVar("true")));
  HashTableInsert(ht, "print.level", param_context,
		  makePermanent(newScalarVar(3.0)));
  HashTableInsert(ht, "print.precision", param_context,
		  makePermanent(newScalarVar(3.0)));
  HashTableInsert(ht, "print.bitspace", param_context,
		  makePermanent(newStringVar("")));
  HashTableInsert(ht, "print.newline", param_context,
		  makePermanent(newStringVar("true")));
  HashTableInsert(ht, "print.separator", param_context,
		  makePermanent(newStringVar("")));
  HashTableInsert(ht, "print.flush", param_context,
		  makePermanent(newStringVar("true")));
  HashTableInsert(ht, "bitvec.threshold", param_context,
		  makePermanent(newScalarVar(0.0)));
  HashTableInsert(ht, "debug.tree", param_context,
		  makePermanent(newStringVar(debugTree ? "true" : "false")));
  HashTableInsert(ht, "debug.addr", param_context,
		  makePermanent(newStringVar(debugAddr ? "true" : "false")));
  HashTableInsert(ht, "debug.alloc", param_context,
		  makePermanent(newStringVar(debugAlloc ? "true" : "false")));
  HashTableInsert(ht, "debug.mem", param_context,
		  makePermanent(newStringVar("false")));
  HashTableInsert(ht, "orthogonalize.precision", param_context,
		  makePermanent(newScalarVar(1e-3)));
  HashTableInsert(ht, "true", param_context,
		  makeConstant(makePermanent(newStringVar("true"))));
  HashTableInsert(ht, "false", param_context,
		  makeConstant(makePermanent(newStringVar("false"))));
  HashTableInsert(ht, "pi", var_context,
		  makeConstant(makePermanent(newScalarVar(M_PI))));

  for (i=0; i<rows; i++) {
    var = CallocOrAbort(struct VarNode, 1);
    var->type = Vector;
    var->d.vec.dim = cols;
    var->d.vec.ptr = M[i];
    var->permanent = 1;
    var->constant = 0;
    var->name = matlabels[i];
    HashTableInsert(ht, matlabels[i], var_context, (void*) var);
  }

  initBitCountTable();

  if (files==NULL)
    yyin = stdin;
  else
    yyin = files[0];
  usePrompt = isatty(fileno(yyin));

  while (!done) {
    eof_flag = 0;
    if (setjmp(abort_expr_env)==0) {
      usePrompt = isatty(fileno(yyin));
      if (debug) fprintf(dout, "Reading next command interactive=%d usePrompt=%d suppressPrompt=%d...\n", isatty(fileno(yyin)), usePrompt, suppressPrompt);
      if ((usePrompt || trueString(stringVarValue("echo"))) && !suppressPrompt)
	fprintf(dout, "> ");
      suppressPrompt = 0;
      expr = ParseExpr();
      if (strlen(inputLine)>0 && inputLine[strlen(inputLine)-1]!='\n')
	suppressPrompt = 1;
      if (trueString(stringVarValue("echo"))) {
	if (strlen(inputLine)>0 && strcmp(inputLine, "\n")) {
	  fprintf(dout, "%s", inputLine);
	  /* should this print newline be outside this if? */
	  if (suppressPrompt)
	    fprintf(dout, "\n");
	}
      }
      if (expr!=NULL) {
	if (trueString(stringVarValue("debug.tree"))) {
	  printTree(expr);
	  fprintf(dout, "\n");
	}
	local_context = var_context+1;
	if (trueString(stringVarValue("debug.addr"))) {
	  fprintf(dout, "Pre evaluation:\n");
	  printDebugTree(expr, 0);
	}
	evalTree(expr);
	if (trueString(stringVarValue("debug.addr"))) {
	  fprintf(dout, "Post evaluation:\n");
	  printDebugTree(expr, 0);
	}
	if (!suppressPrinting(expr)) {
	  if (expr->type==Label)
	    fprintf(dout, "%s ", expr->name);
	  printVarValue(expr->var, (int) scalarVarValue("print.level"));
	  fprintf(dout, "\n");
	  fflush(dout);
	}
	freeNode(expr);
	freeLocalVars();
	freeFreeList();
	debugAlloc = trueString(stringVarValue("debug.alloc"));
#ifdef ARENA_MEM
	if (debugAlloc)
	  amallopt(M_DEBUG, 1, arena);
	else
	  amallopt(M_DEBUG, 0, arena);
#endif
#ifndef ARENA_MEM
	if (debugAlloc) {
	  if (!debugArenaWarned)
	    fprintf(stderr, "debug.alloc (-debugAlloc) has no effect when calc not compiled with -DARENA_MEM\n");
	  debugArenaWarned = 1;
	} else
	  debugArenaWarned = 0;
#endif
      }
      if (eof_flag) {
	if (usePrompt || trueString(stringVarValue("echo")))
	  fprintf(dout, "bye\n");
	done = 1;
      }
    }
  }
  return 0;
}

void initBitCountTable()
{
  unsigned long int x;  /* careful with x, because we use it for loop control,
			   and it has to be able to hold a value of
			   BIT_COUNT_TABLE_SIZE+1 */
  bitWord y;
  int c, i;
#ifdef HAVE_BIT_CHUNK_TYPE
  bitChunk bc;
#endif

  /* printf("BIT_COUNT_CHUNK_SIZE=%d\n", BIT_COUNT_CHUNK_SIZE); */
  /* printf("BIT_COUNT_TABLE_SIZE=%d\n", BIT_COUNT_TABLE_SIZE); */
  /* printf("BIT_COUNT_CHUNK_MASK=%X\n", BIT_COUNT_CHUNK_MASK); */
  /* printf(" BIT_CHUNKS_PER_WORD=%d\n",  BIT_CHUNKS_PER_WORD); */

#ifdef HAVE_BIT_CHUNK_TYPE
  if (BITS(bc)!=BIT_COUNT_CHUNK_SIZE)
    ErrorAbort("bad constants: BITS(BIT_COUNT_CHUNK_TYPE)!=BIT_COUNT_CHUNK_SIZE");
  if (BIT_CHUNKS_PER_WORD*BIT_COUNT_CHUNK_SIZE != BITS(bitWord))
    ErrorAbort("can't use BIT_COUNT_CHUNK_TYPE if it does not evenly go into bitWord");
#endif

  for (x=0; x<BIT_COUNT_TABLE_SIZE; x++) {
    c = 0;
    y = x;
    for (i=0; i<BIT_COUNT_CHUNK_SIZE; i++) {
      if (!(y & 1)) c++;
      y >>= 1;
    }
    bitCountTable[x] = c;
    /* printf("%X %d\n", x, c); */
  }
}

int bitDotProduct(int dim, bitWord *x, bitWord *y)
{
  int n, i, j, c;
  bitWord z;
#ifdef HAVE_BIT_CHUNK_TYPE
  bitChunk *pz = (bitChunk*) &z;
#endif
  n = (dim-1)/BITS(bitWord)+1; /* number of bitWords */
  c = 0;
  /* unroll the inner loop */
  switch (BIT_CHUNKS_PER_WORD) {
  case 1:
    for (i=0; i<n; i++) {
      z = x[i] ^ y[i];
      c += bitCountTable[z];
    }
    break;
  case 2:
#ifdef HAVE_BIT_CHUNK_TYPE
    for (i=0; i<n; i++) {
      z = x[i] ^ y[i];
      c += bitCountTable[pz[0]];
      c += bitCountTable[pz[1]];
    }
#else
    for (i=0; i<n; i++) {
      z = x[i] ^ y[i];
      c += bitCountTable[z & BIT_COUNT_CHUNK_MASK];
      c += bitCountTable[z>>BIT_COUNT_CHUNK_SIZE];
    }
#endif
    break;
  case 4:
#ifdef HAVE_BIT_CHUNK_TYPE
    for (i=0; i<n; i++) {
      z = x[i] ^ y[i];
      c += bitCountTable[pz[0]];
      c += bitCountTable[pz[1]];
      c += bitCountTable[pz[2]];
      c += bitCountTable[pz[3]];
    }
#else
    for (i=0; i<n; i++) {
      z = x[i] ^ y[i];
      c += bitCountTable[z & BIT_COUNT_CHUNK_MASK];
      z >>= BIT_COUNT_CHUNK_SIZE;
      c += bitCountTable[z & BIT_COUNT_CHUNK_MASK];
      z >>= BIT_COUNT_CHUNK_SIZE;
      c += bitCountTable[z & BIT_COUNT_CHUNK_MASK];
      z >>= BIT_COUNT_CHUNK_SIZE;
      c += bitCountTable[z];
    }
#endif
    break;
  default:
    for (i=0; i<n; i++) {
      z = x[i] ^ y[i];
      /* printf("Word %d is %X %X %d %d\n", i, z, z & BIT_COUNT_CHUNK_MASK,
	 z & BIT_COUNT_CHUNK_MASK,
	 bitCountTable[z & BIT_COUNT_CHUNK_MASK]);*/
      j = 0;
      while (1) {
	c += bitCountTable[z & BIT_COUNT_CHUNK_MASK];
	j++;
	if (j>=BIT_CHUNKS_PER_WORD)
	  break;
	z >>= BIT_COUNT_CHUNK_SIZE;
      }
    }
  }
  /* correct for the extra matches in the last unfilled word */
  c -= (n*BITS(bitWord) - dim);
  return c;
}

int min(int x, int y)
{
  return x > y ? y : x;
}

/***
 * massage:
 *   strip quotes off and cope with \n and \"
 */
char *massage(char *input)
{
  char *output = CallocOrAbort(char, strlen(input)-1);
  char *start;
  start = output;
  input++;
  while (*input) {
    if (input[0]=='\\' && input[1]=='n') {
      input+=2;
      *(output++) = '\n';
    } else if (input[0]=='\\' && input[1]=='"') {
      input+=2;
      *(output++) = '"';
    } else if (input[0]=='\\' && input[1]=='t') {
      input+=2;
      *(output++) = '\t';
    } else {
      *(output++) = *(input++);
    }
  }
  /* clobber trailing double-quote */
  if (output > start)
    output[-1] = '\0';
  return start;
}

void catInput(char *input)
{
  while (*input)
    *(inputLineEnd++) = *(input++);
  *inputLineEnd = '\0';
}

void AbortExpr(char * format, ...)
{
  va_list args;
  va_start(args, format);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, format, args);
  fprintf(stderr, "\n");
  va_end(args);
  longjmp(abort_expr_env, 1);
}

/* use a list of things to free, to avoid freeing them twice */
void addToFreeList(void *ptr)
{
  int i;
  for (i=0; i<freeListTop; i++)
    if (freeList[i]==ptr)
      return;
  if (freeListTop==freeListUpper) {
    freeList = ReallocOrAbort(void *, freeList,
			      freeListUpper+FREE_LIST_EXTEND);
    freeListUpper += FREE_LIST_EXTEND;
  }
  freeList[freeListTop] = ptr;
  freeListTop++;
}

void freeFreeList()
{
  int i;
  for (i=freeListTop-1; i>=0; i--)
    free(freeList[i]);
  freeListTop = 0;
}

void freeVar(struct VarNode *var)
{
  if (var==NULL || var->permanent)
    return;
  /* any substructures in this to free? */
  switch (var->type) {
  case Vector:
    /* fprintf(stderr, "Freeing 0x%x (var->d.vec.ptr)\n", var->d.vec.ptr); */
    if (var->d.vec.ptr!=NULL)
      addToFreeList(var->d.vec.ptr);
    break;
  case String:
    /* fprintf(stderr, "Freeing 0x%x (var->d.str)\n", (var->d.str)); */
    if (var->d.str!=NULL)
      addToFreeList(var->d.str);
    break;
  case LNode:
    if (var->d.l.car!=NULL) freeVar(var->d.l.car);
    if (var->d.l.cdr!=NULL) freeVar(var->d.l.cdr);
    break;
  case BitVector:
    if (var->d.vec.bw!=NULL)
      addToFreeList(var->d.vec.bw);
    break;
  default:
    break;
  }
  /* fprintf(stderr, "Freeing 0x%x (var->name)\n", (var->name)); */
  if (var->name!=NULL)
    addToFreeList(var->name);
  /* fprintf(stderr, "Freeing 0x%x (var)\n", (var)); */
  addToFreeList(var);
}

void freeNode(struct ExprNode *node)
{
  if (node==NULL) return;
  freeNode(node->arg1);
  freeNode(node->arg2);
  /* fprintf(stderr, "Freeing 0x%x (node->name)\n", (node->name)); */
  if (node->name!=NULL)
    addToFreeList(node->name);
  freeVar(node->var);
  /* fprintf(stderr, "Freeing 0x%x (node)\n", (node)); */
  addToFreeList(node);
}

void markUnevaluated(struct ExprNode *node)
{
  if (node==NULL) return;
  node->evaluated = 0;
  markUnevaluated(node->arg1);
  markUnevaluated(node->arg2);
  if (node->type!=Literal && node->var!=NULL) {
    if (!node->var->permanent)
      freeVar(node->var);
    node->var = NULL;
  }
}

void freeLocalVars()
{
  int context;
  for (context=local_context-1; context>var_context; context--)
    HashTableDeleteByContext(ht, context);
}

struct ExprNode *newNode
  (enum NodeType type,
   int operator,
   struct ExprNode *arg1,
   struct ExprNode *arg2)
{
  struct ExprNode *node;
  node = CallocOrAbort(struct ExprNode, 1);
  node->type = type;
  node->operator = operator;
  node->arg1 = arg1;
  node->arg2 = arg2;
  node->evaluated = 0;
  node->name = NULL;
  node->var_image = NULL;
  if (node->type==Literal) {
    node->var = CallocOrAbort(struct VarNode, 1);
    node->var->name = NULL;
  } else
    node->var = NULL;
  return node;
}

char *varType(struct VarNode *var)
{
  switch (var->type) {
  case Vector:
    return "vector";
  case Scalar:
    return "scalar";
  case String:
    return "string";
  case LNode:
    return "lnode";
  default:
    AbortExpr("bad var type %d", var->type);
  }
  return "";
}

char *opName(int op)
{
  switch (op) {
  case EQUALS:
    return "=";
  case MINUS:
    return "-";
  case PLUS:
    return "+";
  case TIMES:
    return "*";
  case CTIMES:
    return "@*";
  case DIVIDE:
    return "/";
  case CDIVIDE:
    return "@/";
  case DOT:
    return ".";
  case EXP:
    return "**";
  case HAT:
    return "^";
  case NEG:
    return "-";
  case QUOTE:
    return "'";
  case LANGLE:
    return "<>";
  case BAR:
    return "||";
  case LCURLY:
    return "{}";
  case LSQUARE:
    return "[]";
  default:
    AbortExpr("opName: unknown operator %d", op);
  }
  return "";
}

void printOp(int op)
{
  fprintf(dout, "%s", opName(op));
}

int suppressPrinting(struct ExprNode *node)
{
  if (node==NULL)
    return 1;
  switch (node->type) {
  case Literal:
  case Variable:
  case List:
  case Label:
  case Sequence:
    return 0;
    break;
  case Operator:
    return node->operator == EQUALS;
    break;
  case Function:
    if (   !strcmp(node->name, "store")
	|| !strcmp(node->name, "match")
	|| !strcmp(node->name, "orthogonalize")
	|| !strcmp(node->name, "push")
	|| !strcmp(node->name, "pop")
	|| !strcmp(node->name, "top")
	|| !strcmp(node->name, "cons")
	|| !strcmp(node->name, "repeat")
	|| !strcmp(node->name, "silent")
	|| !strcmp(node->name, "varalias")
	|| !strcmp(node->name, "randvec")
	|| !strcmp(node->name, "print"))
      return 1;
    else
      return 0;
    break;
  default:
    ErrorAbort("suppressPrinting: bad node type: %d", node->type);
  }
  return 1;
}

void printVarValue(struct VarNode *var, int level)
{
  int i, j, k;
  int precision;
  char *space;
  bitWord b;
  switch (var->type) {
  case Vector:
    if (trueString(stringVarValue("print.brace")))
      fprintf(dout, "{");
    precision = (int) scalarVarValue("print.precision");
    for (i=0; i<var->d.vec.dim; i++)
      fprintf(dout, " %.*g"+(i==0?1:0), precision, var->d.vec.ptr[i]);
    if (trueString(stringVarValue("print.brace")))
      fprintf(dout, "}");
    break;
  case BitVector:
    space = stringVarValue("print.bitspace");
    if (strlen(space)==0)
      space = NULL;
    if (trueString(stringVarValue("print.brace")))
      fprintf(dout, "{");
    for (i=0; i<var->d.vec.dim; i++) {
      k = i / BITS(bitWord);
      j = i % BITS(bitWord);
      b = var->d.vec.bw[k];
      if (i>0 && space)
	fprintf(dout, "%s", space);
      fprintf(dout, "%d", (b&(1<<j))!=0);
    }
    if (trueString(stringVarValue("print.brace")))
      fprintf(dout, "}");
    break;
  case Scalar:
    if (rint(var->d.value) == var->d.value) {
      fprintf(dout, "%d", (int) var->d.value);
    } else {
      precision = (int) scalarVarValue("print.precision");
      fprintf(dout, "%.*g", precision, var->d.value);
    }
    break;
  case String:
    fprintf(dout, "%s", var->d.str);
    break;
  case LNode:
    if (level>0) {
      fprintf(dout, "(");
      if (var->d.l.car) 
	printVarValue(var->d.l.car, level-1);
      fprintf(dout, " ");
      if (var->d.l.cdr) 
	printVarValue(var->d.l.cdr, level-1);
      fprintf(dout, ")");
    } else
      fprintf(dout, "( ... )");
    break;
  default:
    AbortExpr("unknown literal type %d\n", var->type);
  }
}

void printTree(struct ExprNode *node)
{
  char *bracket;
  switch (node->type) {
  case Literal:
    printVarValue(node->var, MMAXINT);
    break;
  case Variable:
    fprintf(dout, "%s", node->name);
    break;
  case Function:
    fprintf(dout, "%s(", node->name);
    if (node->arg1)
      printTree(node->arg1);
    if (node->arg2) {
      fprintf(dout, "; ");
      printTree(node->arg2);
    }
    fprintf(dout, ")");
    break;
  case Label:
    fprintf(dout, "\"%s\" ", node->name);
    printTree(node->arg1);
    break;
  case List:
    if (node->arg1)
      printTree(node->arg1);
    if (node->arg2) {
      if (node->operator==COMMA)
	fprintf(dout, ", ");
      else
	fprintf(dout, " ");
      printTree(node->arg2);
    }
    break;
  case Sequence:
    fprintf(dout, "(");
    printTree(node->arg1);
    fprintf(dout, ", ");
    printTree(node->arg2);
    fprintf(dout, ")");
    break;
  case Operator:
    switch (node->operator) {
    case EQUALS:
      if (node->arg1==NULL) {
	printOp(node->operator);
	printTree(node->arg2);
	break;
      }
    case PLUS:
    case MINUS:
    case TIMES:
    case DIVIDE:
    case CTIMES:
    case CDIVIDE:
    case DOT:
    case EXP:
    case HAT:
      fprintf(dout, "(");
      printTree(node->arg1);
      printOp(node->operator);
      printTree(node->arg2);
      fprintf(dout, ")");
      break;
    case NEG:
      printOp(node->operator);
      printTree(node->arg1);
      break;
    case QUOTE:
      printTree(node->arg1);
      printOp(node->operator);
      break;
    case COMMA:
      if (node->arg1->type==COMMA || node->arg2->type==COMMA)
	fprintf(dout, "(");
      if (node->arg1)
	printTree(node->arg1);
      if (node->arg2) {
	fprintf(dout, ", ");
	printTree(node->arg2);
      }
      if (node->arg1->type==COMMA || node->arg2->type==COMMA)
	fprintf(dout, ")");
      break;
    case LPAREN:
    case LSQUARE:
    case LCURLY:
    case BAR:
      bracket = (  node->operator==LANGLE ? "<>"
		 : node->operator==LPAREN ? "()"
		 : node->operator==LSQUARE ? "[]"
		 : node->operator==LCURLY ? "{}"
		 : "||");
      fprintf(dout, "%c", bracket[0]);
      if (node->arg1)
	printTree(node->arg1);
      fprintf(dout, "%c", bracket[1]);
      break;
    case LANGLE:
      fprintf(dout, "<");
      if (node->arg1) {
	printTree(node->arg1);
      }
      if (node->arg2) {
	fprintf(dout, "; ");
	printTree(node->arg2);
      }
      fprintf(dout, ">");
      break;
    default:
      AbortExpr("printTree: bad operator type %d", node->operator);
      break;
    }
    break;
  default:
    AbortExpr("printTree: bad node type %d", node->type);
    break;
  }
}

void printDebugTree(struct ExprNode *node, int depth)
{
  char *bracket;
  int precision;
  if (depth>0) fprintf(dout, "\n");
  fprintf(dout, "0x%-8lx   %*s", (long) node, depth*2, "");
  if (node==NULL) return;
  switch (node->type) {
  case Literal:
    printVarValue(node->var, MMAXINT);
    break;
  case Variable:
    fprintf(dout, "%s", node->name);
    break;
  case Function:
    fprintf(dout, "%s(...)", node->name);
    break;
  case Label:
    fprintf(dout, "\"%s\" ", node->name);
    break;
  case List:
    if (node->arg2) {
      if (node->operator==COMMA)
	fprintf(dout, ", (List)");
      else
	fprintf(dout, "(List)");
    }
    break;
  case Sequence:
    fprintf(dout, ", (Sequence)");
    break;
  case Operator:
    switch (node->operator) {
    case EQUALS:
    case PLUS:
    case MINUS:
    case TIMES:
    case DIVIDE:
    case CTIMES:
    case CDIVIDE:
    case DOT:
    case EXP:
    case HAT:
      printOp(node->operator);
      break;
    case NEG:
      printOp(node->operator);
      break;
    case QUOTE:
      printOp(node->operator);
      break;
    case COMMA:
      if (node->arg1->type==COMMA || node->arg2->type==COMMA)
	fprintf(dout, ", (COMMA Operator)");
      break;
    case LPAREN:
    case LSQUARE:
    case LCURLY:
    case BAR:
    case LANGLE:
      bracket = (  node->operator==LANGLE ? "<>"
		 : node->operator==LPAREN ? "()"
		 : node->operator==LSQUARE ? "[]"
		 : node->operator==LCURLY ? "{}"
		 : "||");
      fprintf(dout, "%s", bracket);
      break;
    default:
      AbortExpr("printDebugTree: bad operator type %d", node->operator);
      break;
    }
    break;
  default:
    AbortExpr("printDebugTree: bad node type %d", node->type);
    break;
  }

  if (depth<15)
    fprintf(dout, "%*s", 30-depth*2, "");
  fprintf(dout, "0x%-8lx", (long) node->var);
  if (node->var!=NULL) {
    fprintf(dout, " %s", node->var->permanent ? "perm" : "temp");
    switch (node->var->type) {
    case BitVector:
      fprintf(dout, " BitVector dim %d", node->var->d.vec.dim);
      break;
    case Vector:
      fprintf(dout, " Vector dim %d", node->var->d.vec.dim);
      break;
    case Scalar:
      precision = (int) scalarVarValue("print.precision");
      fprintf(dout, " Scalar %.*g", precision, node->var->d.value);
      break;
    case String:
      fprintf(dout, " String \"%s\"", node->var->d.str);
      break;
    case LNode:
      fprintf(dout, " LNode (0x%lx, 0x%lx)",
	      (long) node->var->d.l.car, (long) node->var->d.l.cdr);
      break;
    default:
      AbortExpr("unknown literal type %d\n", node->var->type);
    }
  }

  if (node->arg1)
    printDebugTree(node->arg1, depth+1);
  if (node->arg2)
    printDebugTree(node->arg2, depth+1);

  if (depth==0)
    fprintf(dout, "\n");
}

char *ith(int i)
{
  static char buf[10];
  if (i%10 == 1)
    sprintf(buf, "%dst", i);
  else if (i%10 == 2)
    sprintf(buf, "%dnd", i);
  else if (i%10 == 3)
    sprintf(buf, "%drd", i);
  else
    sprintf(buf, "%dth", i);
  return buf;
}

int standardVectorType(int n)
{
  char *str = stringVarValue("vector.type");
  int len = strlen(str);
  if (len>0 && (   strncmp(str, "normal", len)==0
		|| strncmp(str, "standard", len)==0))
    return 1;
  else if (len>0 && (   strncmp(str, "circular", len)==0
		     || strncmp(str, "unitary", len)==0)) {
    if (n%2 != 0)
      AbortExpr("vector dim must be even for circular vectors");
    return 0;
  } else
    AbortExpr("vector.type must be set to \"standard\" or \"circular\"");
  return 1;
}

int trueString(char *str)
{
  if (str==NULL)
    return 0;
  return (   !strncasecmp("true", str, strlen(str))
	  || !strncasecmp("yes", str, strlen(str))
	  || !strncasecmp("1", str, strlen(str)));
}

real *nodeRealDataPtr
  (struct ExprNode *node,
   int *n,
   char *what,	/* function name for error message */
   int i)	/* number of argument */
{
  /***
   * Treat it as a vector of reals
   */
  if (node->var==NULL)
    AbortExpr("%s: missing variable record");
  if (node->var->type==Scalar) {
    *n = 1;
    return &(node->var->d.value);
  } else if (node->var->type==Vector) {
    *n = node->var->d.vec.dim;
    return node->var->d.vec.ptr;
  } else
    AbortExpr("%s argument for \"%s\" must be number or vector",
	      i==0?"":ith(i), what==NULL ? "internal storage" : what);
  return NULL;
}

real *varRealDataPtr
  (struct VarNode *var,
   int *n,
   char *what,	/* function name for error message */
   int i)	/* number of argument */
{
  /***
   * Treat it as a vector of reals
   */
  if (var==NULL)
    AbortExpr("%s: missing variable record");
  if (var->type==Scalar) {
    *n = 1;
    return &(var->d.value);
  } else if (var->type==Vector) {
    *n = var->d.vec.dim;
    return var->d.vec.ptr;
  } else
    AbortExpr("%s argument for \"%s\" must be number or vector",
	      i==0?"":ith(i), what==NULL ? "internal storage" : what);
  return NULL;
}

struct VarNode *lookupVar(char *name)
{
  int context;
  struct HashTableItem *item = NULL;
  for (context=local_context-1; context>=param_context; context--) {
    item = HashTableLookup(ht, name, context);
    if (item!=NULL)
      break;
  }
  if (item==NULL)
    AbortExpr("variable \"%s\" does not exist", name);
  return item->data.ptr;
}

struct VarNode *scalarVar(char *name)
{
  struct VarNode *var = lookupVar(name);
  if (var->type!=Scalar)
    AbortExpr("variable \"%s\" is not a scalar", name);
  return var;
}

real scalarVarValue(char *name)
{
  struct VarNode *var = scalarVar(name);
  return var->d.value;
}

char *stringVarValue(char *name)
{
  struct VarNode *var = lookupVar(name);
  if (var->type!=String)
    AbortExpr("variable \"%s\" is not a string", name);
  return var->d.str;
}

int setScalarVar(char *name, Real value)
{
  struct VarNode *var = lookupVar(name);
  if (var->type!=Scalar)
    AbortExpr("variable \"%s\" is not a scalar");
  var->d.value = value;
  return 1;
}

int setStringVar(char *name, char *value)
{
  struct VarNode *var = lookupVar(name);
  if (var->type!=String)
    AbortExpr("variable \"%s\" is not a string", name);
  var->d.str = strdup(value);
  return 1;
}

struct VarNode *makePermanent(struct VarNode *var)
{
  if (var==NULL) return NULL;
  var->permanent = 1;
  if (var->type == LNode) {
    makePermanent(var->d.l.car);
    makePermanent(var->d.l.cdr);
  }
  return var;
}

struct VarNode *makeConstant(struct VarNode *var)
{
  if (var==NULL) return NULL;
  var->constant = 1;
  if (var->type == LNode) {
    makeConstant(var->d.l.car);
    makeConstant(var->d.l.cdr);
  }
  return var;
}

struct VarNode *newScalarVar(real x)
{
  struct VarNode *var = CallocOrAbort(struct VarNode, 1);
  var->type = Scalar;
  var->permanent = 0;
  var->constant = 0;
  var->d.value = x;
  var->name = NULL;
  return var;
}

struct VarNode *newStringVar(char *str)
{
  struct VarNode *var = CallocOrAbort(struct VarNode, 1);
  var->type = String;
  var->permanent = 0;
  var->constant = 0;
  var->d.str = strdup(str);
  var->name = NULL;
  return var;
}

struct VarNode *newVectorVar(int n, real *v)
{
  int i;
  struct VarNode *var = CallocOrAbort(struct VarNode, 1);
  var->type = Vector;
  var->permanent = 0;
  var->constant = 0;
  if (n>0)
    var->d.vec.ptr = CallocOrAbort(real, n);
  else
    var->d.vec.ptr = 0;
  var->d.vec.dim = n;
  var->name = NULL;
  if (v!=NULL)
    for (i=0; i<n; i++)
      var->d.vec.ptr[i] = v[i];
  return var;
}

struct VarNode *newBitVectorVar(int n, bitWord *b)
{
  int i, words;
  struct VarNode *var = CallocOrAbort(struct VarNode, 1);
  var->type = BitVector;
  var->permanent = 0;
  var->constant = 0;
  words = (n+BITS(bitWord)-1)/BITS(bitWord);
  var->d.vec.dim = n;
  var->d.vec.bw = CallocOrAbort(bitWord, words);
  var->name = NULL;
  if (b!=NULL)
    for (i=0; i<words; i++)
      var->d.vec.bw[i] = b[i];
  return var;
}

struct VarNode *newLNodeVar(struct VarNode *car, struct VarNode *cdr)
{
  struct VarNode *var = CallocOrAbort(struct VarNode, 1);
  var->type = LNode;
  var->permanent = 0;
  var->constant = 0;
  var->d.l.car = car;
  var->d.l.cdr = cdr;
  var->name = NULL;
  return var;
}

struct VarNode *dupVar(struct VarNode *image)
{
  struct VarNode *var = CallocOrAbort(struct VarNode, 1);
  var->type = image->type;
  var->permanent = 0;
  var->constant = 0;
  var->name = NULL;
  switch (var->type) {
  case Vector:
    var->d.vec.dim = image->d.vec.dim;
    if (var->d.vec.dim>0)
      var->d.vec.ptr = CallocOrAbort(real, image->d.vec.dim);
    else
      var->d.vec.ptr = NULL;
    break;
  case Scalar:
    var->d.value = 0.0;
    break;
  case String:
    var->d.str = NULL;
    break;
  case LNode:
    var->d.l.car = NULL;
    var->d.l.cdr = NULL;
    break;
  default:
    ErrorAbort("dupVar: bad image node type %d", var->type);
  }
  return var;
}

struct VarNode *copyVar(struct VarNode *image)
{
  int i;
  struct VarNode *var;
  if (image==NULL) return NULL;
  var = CallocOrAbort(struct VarNode, 1);
  var->type = image->type;
  var->permanent = 0;
  var->constant = 0;
  var->name = NULL;
  switch (var->type) {
  case Vector:
    var->d.vec.dim = image->d.vec.dim;
    var->d.vec.ptr = CallocOrAbort(real, image->d.vec.dim);
    for (i=0; i<var->d.vec.dim; i++)
      var->d.vec.ptr[i] = image->d.vec.ptr[i];
    break;
  case Scalar:
    var->d.value = image->d.value;
    break;
  case String:
    var->d.str = image->d.str!=NULL ? strdup(image->d.str) : NULL;
    break;
  case LNode:
    var->d.l.car = image->d.l.car!=NULL ? copyVar(image->d.l.car) : NULL;
    var->d.l.cdr = image->d.l.cdr!=NULL ? copyVar(image->d.l.cdr) : NULL;
    break;
  default:
    ErrorAbort("copyVar: bad image node type %d", var->type);
  }
  return var;
}

void opNotImplemented
  (int op,
   struct VarNode *var1,
   struct VarNode *var2,
   int swapped)
{
  AbortExpr("Operation (%s) %s (%s) not implemented",
	    varType(swapped?var2:var1),
	    opName(op), varType(swapped?var1:var2));
}

int countListArgs(struct ExprNode *arg)
{
  int nargs;
  if (arg==NULL)
    return 0;
  nargs = 1;
  while (arg->type==List && arg->arg2!=NULL) {
    nargs++;
    arg = arg->arg2;
  }
  return nargs;
}

real sumCosDiff(int n, real *x, real *y)
{
  double sum = 0.0;
  int i;
  for (i=0; i<n; i+=2) {
    if (x[i]!=1.0 || y[i]!=1.0)
      sum += x[i]*y[i]*cos(x[i+1] - y[i+1]);
    else
      sum += cos(x[i+1] - y[i+1]);
  }
  sum /= n/2;
  return sum;
}

void standardVectorArith(int op, int nr, real *rpr, real *rp1, real *rp2,
		       struct ExprNode *arg1, struct ExprNode *arg2)
{
  int i;
  if (trueString(stringVarValue("conv.fft"))
      && (op==TIMES || op==DIVIDE)) {
    for (i=nr; i>0 && (i&1) == 0; i>>=1) {};
    if (i!=1)
      AbortExpr("vector dim must be a power of 2 for conv/corr by fft");
    if (nr>4096)
      AbortExpr("vector dim too large for conv/corr by fft: ffts don't work with n>4096");
  }
  switch (op) {
  case TIMES:
    if (trueString(stringVarValue("conv.fft")))
      convolveFFT(nr, rpr, rp1, rp2);
    else
      convolve(nr, rpr, rp1, rp2);
    break;
  case DIVIDE:
    if (trueString(stringVarValue("conv.fft")))
      correlateFFT(nr, rpr, rp2, rp1);
    else
      correlate(nr, rpr, rp2, rp1);
    break;
  case PLUS:
    for (i=0; i<nr; i++)
      rpr[i] = rp1[i] + rp2[i];
    break;
  case MINUS:
    for (i=0; i<nr; i++)
      rpr[i] = rp1[i] - rp2[i];
    break;
  case CTIMES:
    for (i=0; i<nr; i++)
      rpr[i] = rp1[i] * rp2[i];
    break;
  case CDIVIDE:
    for (i=0; i<nr; i++)
      rpr[i] = rp1[i] / rp2[i];
    break;
  default:
    opNotImplemented(op, arg1->var, arg2->var, 0);
  }
}

void standardScalarVectorArith(int op, int nr,
			     real *rpr, real *rp1, real *rp2,
			     struct ExprNode *arg1, struct ExprNode *arg2,
			     int swapped)
{
  int i;
  real x = rp1[0];
  switch (op) {
  case PLUS:
    for (i=0; i<nr; i++)
      rpr[i] = x + rp2[i];
    break;
  case MINUS:
    for (i=0; i<nr; i++)
      rpr[i] = x - rp2[i];
    break;
  case TIMES:
  case CTIMES:
    for (i=0; i<nr; i++)
      rpr[i] = x * rp2[i];
    break;
  case DIVIDE:
  case CDIVIDE:
    if (!swapped)
      for (i=0; i<nr; i++)
	rpr[i] = x / rp2[i];
    else
      for (i=0; i<nr; i++)
	rpr[i] = rp2[i] / x;
    break;
  case EXP:
  case HAT:
    if (trueString(stringVarValue("conv.fft"))) {
      for (i=nr; i>0 && (i&1) == 0; i>>=1) {};
      if (i!=1)
	AbortExpr("vector dim must be a power of 2 for conv/corr by fft");
      if (nr>4096)
	AbortExpr("vector dim too large for conv/corr by fft: ffts don't work with n>4096");
      convolvePow(nr, rpr, rp2, x);
    } else
      AbortExpr("can only use convolution powers when conv.fft=true");
    break;
  default:
    opNotImplemented(op, arg1->var, arg2->var, swapped);
  }
}

void standardScalarArith(int op, real *rpr, real *rp1, real *rp2,
		       struct ExprNode *arg1, struct ExprNode *arg2)
{
  switch (op) {
  case PLUS:
    rpr[0] = rp1[0] + rp2[0];
    break;
  case MINUS:
    rpr[0] = rp1[0] - rp2[0];
    break;
  case TIMES:
  case CTIMES:
    rpr[0] = rp1[0] * rp2[0];
    break;
  case DIVIDE:
  case CDIVIDE:
    rpr[0] = rp1[0] / rp2[0];
    break;
  case EXP:
  case HAT:
    rpr[0] = pow(rp1[0], rp2[0]);
    break;
  default:
    opNotImplemented(op, arg1->var, arg2->var, 0);
  }
}

void circularVectorArith(int op, int nr, real *rpr, real *rp1, real *rp2,
			 struct ExprNode *arg1, struct ExprNode *arg2)
{
  /* only have to do mult & div here - add & sub handled elsewhere */
  int i;
  switch (op) {
  case TIMES:
    for (i=0; i+1<nr; i+=2) {
      rpr[i+1] = rp1[i+1] + rp2[i+1];
      if (rp1[i]==1.0 && rp2[i]==1.0)
	rpr[i] = 1.0;
      else {
	rpr[i] = rp1[i] * rp2[i];
	if (rpr[i]<0) {
	  rpr[i] = -rpr[i];
	  rpr[i+1] += M_PI;
	}
      }
      if (rpr[i+1]>M_PI)
	rpr[i+1] = fmod(rpr[i+1]+M_PI, 2*M_PI)-M_PI;
      else if (rpr[i+1]<-M_PI)
	rpr[i+1] = -fmod(-rpr[i+1]+M_PI, 2*M_PI)+M_PI;
    }
    break;
  case DIVIDE:
    for (i=0; i+1<nr; i+=2) {
      rpr[i+1] = rp1[i+1] - rp2[i+1];
      if (rp1[i]==1.0 && rp2[i]==1.0)
	rpr[i] = 1.0;
      else {
	rpr[i] = rp1[i] / rp2[i];
	if (rpr[i]<0) {
	  rpr[i] = -rpr[i];
	  rpr[i+1] += M_PI;
	}
      }
      if (rpr[i+1]>M_PI)
	rpr[i+1] = fmod(rpr[i+1]+M_PI, 2*M_PI)-M_PI;
      else if (rpr[i+1]<-M_PI)
	rpr[i+1] = -fmod(-rpr[i+1]+M_PI, 2*M_PI)+M_PI;
    }
    break;
  default:
    AbortExpr("operator \"%s\" not implemented for circular vectors",
	      opName(op));
  }
}

#define NormalizeComplex(x, i) {\
      if (x[i]<0) {\
	x[i] = -x[i];\
	x[i+1] += M_PI;\
      }\
      if (x[i+1]>M_PI)\
	x[i+1] = fmod(x[i+1]+M_PI, 2*M_PI)-M_PI;\
      else if (x[i+1]<-M_PI)\
	x[i+1] = -fmod(-x[i+1]+M_PI, 2*M_PI)+M_PI;\
}

void circularScalarVectorArith(int op, int nr,
			       real *rpr, real *rp1, real *rp2,
			       struct ExprNode *arg1, struct ExprNode *arg2,
			       int swapped)
{
  /* only have to do mult, div & exp here - add & sub handled elsewhere */
  int i;
  real x = rp1[0];
  switch (op) {
  case TIMES:
    /* rpr[i+1] is the angle */
    /* rpr[i] is the magnitude */
    for (i=0; i+1<nr; i+=2) {
      rpr[i] = rp2[i] * x;
      rpr[i+1] = rp2[i+1];
      NormalizeComplex(rpr, i);
    }
    break;
  case DIVIDE:
    for (i=0; i+1<nr; i+=2) {
      if (!swapped) {
	rpr[i] = x/rp2[i];
	rpr[i+1] = -rp2[i+1];
      } else {
	rpr[i] = rp2[i]/x;
	rpr[i+1] = rp2[i+1];
      }
      NormalizeComplex(rpr, i);
    }
    break;
  case EXP:
  case HAT:
    for (i=0; i+1<nr; i+=2) {
      rpr[i] = (rp2[i]==1.0 ? 1.0 : pow(rp2[i], x));
      rpr[i+1] = rp2[i+1];
      /*
       * do interval modulus before and after angle multiplication
       * to avoid strange results on vectors not in proper format
       */
      NormalizeComplex(rpr, i);
      rpr[i+1] *= x;
      NormalizeComplex(rpr, i);
    }
    break;
  default:
    AbortExpr("operator \"%s\" not implemented for circular vectors",
	      opName(op));
  }
}

/* circularScalarVectorArithOld -- old version which did scalar-vector
   multiplication by adding the scalar to the angle */

void circularScalarVectorArithOld(int op, int nr,
			       real *rpr, real *rp1, real *rp2,
			       struct ExprNode *arg1, struct ExprNode *arg2,
			       int swapped)
{
  /* only have to do mult, div & exp here - add & sub handled elsewhere */
  int i;
  real x = rp1[0];
  switch (op) {
  case TIMES:
    /* rpr[i+1] is the angle */
    /* rpr[i] is the magnitude */
    for (i=0; i+1<nr; i+=2) {
      rpr[i+1] = x + rp2[i+1];
      rpr[i] = rp2[i];
      if (rpr[i]<0) {
	rpr[i] = -rpr[i];
	rpr[i+1] += M_PI;
      }
      if (rpr[i+1]>M_PI)
	rpr[i+1] = fmod(rpr[i+1]+M_PI, 2*M_PI)-M_PI;
      else if (rpr[i+1]<-M_PI)
	rpr[i+1] = -fmod(-rpr[i+1]+M_PI, 2*M_PI)+M_PI;
    }
    break;
  case DIVIDE:
    for (i=0; i+1<nr; i+=2) {
      if (!swapped) {
	rpr[i+1] = x - rp2[i+1];
	rpr[i] = 1/rp2[i];
      } else {
	rpr[i+1] = rp2[i+1] - x;
	rpr[i] = rp2[i];
      }
      if (rpr[i]<0) {
	rpr[i] = -rpr[i];
	rpr[i+1] += M_PI;
      }
      if (rpr[i+1]>M_PI)
	rpr[i+1] = fmod(rpr[i+1]+M_PI, 2*M_PI)-M_PI;
      else if (rpr[i+1]<-M_PI)
	rpr[i+1] = -fmod(-rpr[i+1]+M_PI, 2*M_PI)+M_PI;
    }
    break;
  case EXP:
  case HAT:
    for (i=0; i+1<nr; i+=2) {
      rpr[i] = pow(rp2[i], x);
      if (rpr[i]<0) {
	rpr[i+1] = rp2[i+1]+M_PI;
	rpr[i] = -rpr[i];
      } else {
	rpr[i+1] = rp2[i+1];
      }
      /*
       * do interval modulus before and after angle multiplication
       * to avoid strange results on vectors not in proper format
       */
      if (rpr[i+1]>M_PI)
	rpr[i+1] = fmod(rpr[i+1]+M_PI, 2*M_PI)-M_PI;
      else if (rpr[i+1]<-M_PI)
	rpr[i+1] = -fmod(-rpr[i+1]+M_PI, 2*M_PI)+M_PI;
      rpr[i+1] *= x;
      if (rpr[i+1]>M_PI)
	rpr[i+1] = fmod(rpr[i+1]+M_PI, 2*M_PI)-M_PI;
      else if (rpr[i+1]<-M_PI)
	rpr[i+1] = -fmod(-rpr[i+1]+M_PI, 2*M_PI)+M_PI;
    }
    break;
  default:
    AbortExpr("operator \"%s\" not implemented for circular vectors", opName(op));
  }
}

void preEvalSummation(struct ExprNode *node, int *isVector,
		      int *count, int *dim, real *s, real **p)
{
  if (node->arg1->type==Operator
      && (node->arg1->operator==PLUS || node->arg1->operator==MINUS))
    preEvalSummation(node->arg1, isVector, count, dim, s, p);
  else {
    evalTree(node->arg1);
    if (node->arg1->var->type==Vector)
      *isVector = 1;
    if (p!=NULL && dim!=NULL)
      p[*count] = nodeRealDataPtr(node->arg1, &(dim[*count]),
				 opName(node->operator), 1);
    if (s!=NULL)
      s[*count] = 1.0;
    (*count)++;
  }
  evalTree(node->arg2);
  if (node->arg2->var->type==Vector)
    *isVector = 1;
  if (p!=NULL && dim!=NULL)
    p[*count] = nodeRealDataPtr(node->arg2, &(dim[*count]),
				opName(node->operator), 2);
  if (s!=NULL)
    s[*count] = (node->operator==PLUS ? 1.0 : -1.0);
  (*count)++;
}

/* circularCOG: circular center of gravity superposition */

void circularCOG(int n, int m, int *dim, real *s, real **p, real *r)
/* n: vector dimension */
/* m: number of vectors */
/* dim[i]: dimension of p[i] */
/* s: scalar factor */
/* p: vector i */
/* r: result vector */
{
  int i, j;
  double h, theta;

  /***
   * Vectors are stored in (r,theta) form in p[j].
   * Start off using rectangular format for r,
   * (x's in r[even] and y's in r[odd]) then
   * convert to polar at end.
   */
  for (i=0; i<n; i++)
    r[i] = 0;
  /* accumlate weighted x's and y's */
  for (j=0; j<m; j++) {
    if (dim==NULL || dim[j]==n) {
      if (s[j]==1.0)
	for (i=0; i+1<n; i+=2) {
	  r[i] += cos(p[j][i+1])*p[j][i];
	  r[i+1] += sin(p[j][i+1])*p[j][i];
	}
      else if (s[j]==-1)
	for (i=0; i+1<n; i+=2) {
	  r[i] -= cos(p[j][i+1])*p[j][i];
	  r[i+1] -= sin(p[j][i+1])*p[j][i];
	}
      else if (s[j]!=0.0)
	for (i=0; i+1<n; i+=2) {
	  r[i] += cos(p[j][i+1])*p[j][i]*s[j];
	  r[i+1] += sin(p[j][i+1])*p[j][i]*s[j];
	}
    } else if (dim[j]==1) {
      /* p[j] is a scalar */
      double x = cos(p[j][0])*s[j];
      double y = sin(p[j][0])*s[j];
      for (i=0; i+1<n; i+=2) {
	r[i] += x;
	r[i+1] += y;
      }
    } else
      AbortExpr("circularCOG: vector dim must be 1 or n");
  }
  /* convert to polar format */
  for (i=0; i+1<n; i+=2) {
    h = hypot(r[i], r[i+1]);
    theta = atan2(r[i+1], r[i]);
    r[i] = h;
    r[i+1] = theta;
  }
}

void evalSummation(struct ExprNode *node, int length)
{
  int *dim = CallocOrAbort(int, length);    /* vector dimensions */
  real **p = CallocOrAbort(real *, length); /* pointers to vectors */
  real *s = CallocOrAbort(real, length);    /* scalar factors (1/-1) */
  int isVector = 0;
  int i, n = 0;
  length = 0;
  preEvalSummation(node, &isVector, &length, dim, s, p);
  for (i=0; i<length; i++) {
    if (dim[i]>1 && dim[i]>n) {
      if (n>0) {
	if (dim[i]!=n)
	  AbortExpr("all vectors in summation must be same dimension");
      } else {
	n = dim[i];
      }
    }
  }
  node->var = newVectorVar(n, NULL);
  circularCOG(n, length, dim, s, p, node->var->d.vec.ptr);
  node->evaluated = 1;
}

void evalTree(struct ExprNode *node)
{
  struct ExprNode *arg1 = node->arg1;
  struct ExprNode *arg2 = node->arg2;
  struct HashTableItem *item = NULL;
  real *rpr, *rp1, *rp2;	/* real pointers for result, arg1 & arg2 */
  int nr, n1, n2;		/* counters for result, arg1, & arg2 */
  int i;
  int swapped = 0;
  double sum = 0;
  int isStandardVectorType = standardVectorType(0);
  if (node->evaluated)
    return;
  switch (node->type) {
  case Literal:
    /***
     * Literal valued token - don't need to do anything
     */
    break;

  case Variable:
    /***
     * Variable symbol - find the variable record from the hash table
     */
    {
      int context;
      for (context=local_context-1; context>=param_context; context--) {
	item = HashTableLookup(ht, node->name, context);
	if (item!=NULL)
	  break;
      }

      if (item!=NULL) {
	node->var = item->data.ptr;
      } else if (node->var_image!=NULL) {
	node->var = dupVar(node->var_image);
	item = HashTableInsert(ht, node->name, var_context, node->var);
	if (item==NULL)
	  AbortExpr("could not create object named \"%s\"", node->name);
	else {
	  node->var->name = item->name;
	  node->var->permanent = 1;
	}
      } else
	AbortExpr("no object named \"%s\"", node->name);
    }
    break;

  case Label:
    evalTree(arg1);
    node->var = arg1->var;
    arg1->var = NULL;
    break;

  case List:
    AbortExpr("can't evaluate a list");
    break;

  case Sequence:
    evalTree(arg1);
    evalTree(arg2);
    node->var = arg2->var;
    arg2->var = NULL;
    break;

  case Function:
    evalFunction(node, node->name);
    break;

  case Operator:
    switch (node->operator) {
      
    case EQUALS:
      doAssignment(node);
      break;

    case PLUS:
    case MINUS:
      /* must do a non-bracketted set of vector additions all at once */
      if (!isStandardVectorType) {
	int isVector = 0;
	int length = 0;
	preEvalSummation(node, &isVector, &length, NULL, NULL, NULL);
	if (isVector) {
	  evalSummation(node, length);
	  break;
	}
      }
      /* else fall through into cases below */
    case TIMES:
    case DIVIDE:
    case CTIMES:
    case CDIVIDE:
    case EXP:
    case HAT:
      evalTree(arg1);
      evalTree(arg2);
      /* make arg1 the scalar */
      if (arg1->var->type==Vector && arg2->var->type==Scalar) {
	arg1 = node->arg2;
	arg2 = node->arg1;
	swapped = 1;
      }
      if ((node->operator==EXP || node->operator==HAT)
	  && arg2->var->type==Vector && !swapped)
	opNotImplemented(node->operator, arg1->var, arg2->var, swapped);
      rp1 = nodeRealDataPtr(arg1, &n1, opName(node->operator), 1);
      rp2 = nodeRealDataPtr(arg2, &n2, opName(node->operator), 2);
      if (n1!=1 && n1!=n2)
	AbortExpr("vector ops require same length vectors");
      node->var = dupVar(arg2->var);
      rpr = nodeRealDataPtr(node, &nr, NULL, 0);

      if (arg1->var->type==Vector && arg2->var->type==Vector) {
	if (standardVectorType(nr))
	  standardVectorArith(node->operator, nr, rpr, rp1, rp2, arg1, arg2);
	else
	  circularVectorArith(node->operator, nr, rpr, rp1, rp2, arg1, arg2);
      } else if (arg1->var->type==Scalar && arg2->var->type==Vector) {
	if (standardVectorType(nr))
	  standardScalarVectorArith(node->operator, nr, rpr, rp1, rp2,
				  arg1, arg2, swapped);
	else
	  circularScalarVectorArith(node->operator, nr, rpr, rp1, rp2,
				    arg1, arg2, swapped);
      } else if (arg1->var->type==Scalar && arg2->var->type==Scalar) {
	standardScalarArith(node->operator, rpr, rp1, rp2, arg1, arg2);
      } else
	AbortExpr("evalTree: internal error: op muck up");

      break;
      
    case DOT:
      evalTree(arg1);
      evalTree(arg2);
      if (arg1->var->type==Vector && arg2->var->type==Vector) {
	rp1 = nodeRealDataPtr(arg1, &n1, opName(node->operator), 1);
	rp2 = nodeRealDataPtr(arg2, &n2, opName(node->operator), 2);
	if (n1!=n2)
	  AbortExpr("%s require same length vectors", opName(DOT));
	if (standardVectorType(n1))
	  sum = dotProduct(n1, rp1, rp2);
	else
	  sum = sumCosDiff(n1, rp1, rp2);
	node->var = newScalarVar(sum);
      } else if (arg1->var->type==BitVector && arg2->var->type==BitVector) {
	bitWord *bw1, *bw2;
	n1 = arg1->var->d.vec.dim;
	bw1 = arg1->var->d.vec.bw;
	n2 = arg2->var->d.vec.dim;
	bw2 = arg2->var->d.vec.bw;
	if (n1!=n2)
	  AbortExpr("%s require same length vectors", opName(DOT));
	node->var = newScalarVar(bitDotProduct(n1, bw1, bw2));
      } else {
	AbortExpr("dot product requires two vectors");
      }
      break;
      
    case LANGLE:
      evalFunction(node, "normalize");
      break;
      
    case NEG:
    case QUOTE:
      evalTree(arg1);
      node->var = dupVar(arg1->var);
      rp1 = nodeRealDataPtr(arg1, &n1, opName(node->operator), 1);
      rpr = nodeRealDataPtr(node, &nr, NULL, 0);
      if (nr==1 || standardVectorType(nr)) {
	switch(node->operator) {
	case NEG:
	  for (i=0; i<nr; i++)
	    rpr[i] = -rp1[i];
	  break;
	case QUOTE:
	  /* convolvative psuedo inverse
	     - reorder the vector [0 n-1 n-2 ... 2 1] */
	  rpr[0] = rp1[0];
	  for (i=1; i<nr; i++)
	    rpr[i] = rp1[nr-i];
	  break;
	}
      } else {
	/* vectors are (r,theta) pairs */
	switch(node->operator) {
	case NEG:
	  for (i=0; i<nr; i+=2) {
	    rpr[i] = rp1[i];
	    if (rp1[i+1]<0)
	      rpr[i+1] = M_PI + rp1[i+1];
	    else if (rp1[i+1]>0)
	      rpr[i+1] = -M_PI + rp1[i+1];
	    else
	      rpr[i+1] = rp1[i+1];
	  }
	  break;
	case QUOTE:
	  /* convolutive psuedo inverse - negate theta */
	  for (i=0; i<nr; i+=2) {
	    rpr[i] = rp1[i];
	    rpr[i+1] = -rp1[i+1];
	  }
	  break;
	}
      }
      break;
      
    case COMMA:
      AbortExpr("comma operator shouldn't be evaluated");
      break;
      
    case BAR:
      evalTree(arg1);
      rp1 = nodeRealDataPtr(arg1, &n1, opName(node->operator), 1);
      sum = 0;
      if (isStandardVectorType)
	for (i=0; i<n1; i++)
	  sum += rp1[i]*rp1[i];
      else {
	for (i=0; i<n1; i+=2)
	  sum += rp1[i]*rp1[i];
	sum /= n1/2;
      }
      node->var = newScalarVar(sqrt(sum));
      break;
      
    case LPAREN:
      evalTree(arg1);
      node->var = arg1->var;
      arg1->var = NULL;
      break;
      
    case LSQUARE:
      AbortExpr("can't evaluate LSQUARE yet");
      break;

    case LCURLY:
      {
	int nargs = 0;
	while (arg1!=NULL) {
	  if (arg1->type!=List)
	    AbortExpr("LCURLY: bad expr structure: need List");
	  if (arg1->arg1==NULL)
	    AbortExpr("LCURLY: bad expr structure: need arg1");
	  evalTree(arg1->arg1);
	  rp1 = nodeRealDataPtr(arg1->arg1, &n1, opName(node->operator), 1);
	  nargs += n1;
	  arg1 = arg1->arg2;
	}
	if (nargs==0)
	  AbortExpr("LCURLY: empty vector");
	node->var = newVectorVar(nargs, NULL);
	nargs = 0;
	rpr = nodeRealDataPtr(node, &nr, opName(node->operator), 1);
	for (arg1=node->arg1; arg1!=NULL; arg1=arg1->arg2) {
	  rp1 = nodeRealDataPtr(arg1->arg1, &n1, opName(node->operator), 1);
	  for (i=0; i<n1; i++)
	    rpr[nargs+i] = rp1[i];
	  nargs += n1;
	}
      }
      break;
      
    default:
      AbortExpr("bad operation type %d", node->operator);
      break;
    }
    break;
  default:
    AbortExpr("bad node type %d", node->type);
    break;
  }
  node->evaluated = 1;
}

void doAssignment(struct ExprNode *node)
{
  struct ExprNode *arg1 = node->arg1;
  struct ExprNode *arg2 = node->arg2;
  int i, nr, n2;
  real *rpr, *rp2;
  evalTree(arg2);
  if (arg1==NULL) {
    node->var = arg2->var;
    return;
  }
  if (arg1->var==NULL) {
    arg1->var_image = arg2->var;
    evalTree(arg1);
    arg1->var_image = NULL;
  }
  if (arg1->type!=Variable)
    AbortExpr("\"%s\" requires variable on left", opName(EQUALS));
  if (arg1->var->constant)
    AbortExpr("cannot change value of \"%s\"", arg1->name);
  node->var = arg1->var;
  arg1->var = NULL;
  if (node->arg2->var->type!=node->var->type)
    AbortExpr("\"%s\" requires same type for each arg", opName(EQUALS));
  if (node->var->type==Vector || node->var->type==Scalar) {
    rp2 = nodeRealDataPtr(node->arg2, &n2, "=", 2);
    rpr = nodeRealDataPtr(node, &nr, NULL, 0);
    if (n2!=nr)
      AbortExpr("vector ops require same length vectors");
    for (i=0; i<nr; i++)
      rpr[i] = rp2[i];
  } else if (node->var->type==String) {
    if (node->var->d.str!=NULL)
      addToFreeList(node->var->d.str);
    node->var->d.str = strdup(node->arg2->var->d.str);
  } else if (node->var->type==LNode) {
    node->var->d.l.car = copyVar(arg2->var->d.l.car);
    node->var->d.l.cdr = copyVar(arg2->var->d.l.cdr);
    if (node->var->permanent) {
      makePermanent(node->var->d.l.car);
      makePermanent(node->var->d.l.cdr);
    }
  }
}

struct VarNode *newBitVectorFromVector(struct VarNode *vec)
{
  struct VarNode *bitvec;
  real threshold;
  real *v;
  int i, j, k, dim;
  bitWord b, m;
  if (vec==NULL || vec->type!=Vector)
    AbortExpr("newBitVectorFromVector: not a vector");
  v = vec->d.vec.ptr;
  dim = vec->d.vec.dim;
  bitvec = newBitVectorVar(dim, NULL);
  threshold = scalarVarValue("bitvec.threshold");
  /* printf("converting vector dim %d thresh %g\n", dim, threshold); */

  j = 0; /* count bits in each bitWord */
  k = 0; /* count bitWords */
  b = 0;
  m = 1;
  for (i=0; i<dim; i++) {
    if (v[i]>threshold)
      b |= m;
    j++;
    m <<= 1;
    if (j>=BITS(bitWord)) {
      bitvec->d.vec.bw[k] = b;
      j = 0;
      b = 0;
      m = 1;
      k++;
    }
  }
  if (j!=0)
    bitvec->d.vec.bw[k] = b;
  return bitvec;
}

void evalFunction(struct ExprNode *node, char *funcName)
{
  int nargs, funcNameLen, id;
  char *localName, *fixedName;
  HashTableItem *item;
  int context = local_context;
  struct ExprNode *localArg = node->arg2;
  local_context++;

  funcNameLen = strlen(funcName);
  while (localArg!=NULL && localArg->type==List) {
    if (localArg->arg1->type!=Operator || localArg->arg1->operator!=EQUALS
	|| localArg->arg1->arg1==NULL || localArg->arg1->arg1->type!=Variable)
      AbortExpr("arguments after \";\" must be var assignments");
    localName = localArg->arg1->arg1->name;
    /* make sure localName is prefixed by the function name */
    if (!strncmp(funcName, localName, funcNameLen)
	&& localName[funcNameLen]=='.') {
      /* it's OK */
      fixedName = NULL;
    } else {
      fixedName = CallocOrAbort(char, strlen(localName)+funcNameLen+2);
      sprintf(fixedName, "%s.%s", funcName, localName);
      localName = fixedName;
    }
    /* check that the name is a local to this func */
    /* look up this var, create a new one in its image in the local context,
       then go ahead and do the assignment */

    item = HashTableLookup(ht, localName, param_context);
    if (item==NULL) {
      fprintf(stderr, "Paramters for this function are:\n");
      id = 0;
      while (1) {
	item = HashTableNextByContext(ht, &id, param_context);
	if (item==NULL) break;
	if (!strncmp(item->name, localName, funcNameLen+1))
	  fprintf(stderr, "  %s\n", item->name);
      }
      AbortExpr("No function parameter named \"%s\".", localName);
    }
    localArg->arg1->arg1->var_image = item->data.ptr;
    if (localArg->arg1->arg1->var_image==NULL)
      AbortExpr("No var associated with \"%s\"", item->name);

    /* look it up at the local context level */
    item = HashTableLookup(ht, localName, context);
    
    if (item!=NULL) {
      localArg->arg1->arg1->var = item->data.ptr;
    } else {
      localArg->arg1->arg1->var = dupVar(localArg->arg1->arg1->var_image);
      item = HashTableInsert(ht, localName, context, localArg->arg1->arg1->var);
      if (item==NULL)
	AbortExpr("could not create object named \"%s\"", localName);
    }
    addToFreeList(localName);
    localArg->arg1->arg1->var_image = NULL;

    /* and now do the assignment -- the local we just created
       should be picked up */
    doAssignment(localArg->arg1);
    /* disconnect the var here to avoid freeing problems */
    localArg->arg1->arg1->var = NULL;

    /* go onto the next local arg */
    localArg = localArg->arg2;
  }

  nargs = countListArgs(node->arg1);
  /* fprintf(dout, "function %s, %d arguments\n", funcName, nargs); */
  if (!strcmp(funcName, "store"))
    node->var = func_store(node->arg1);
  else if (!strcmp(funcName, "match"))
    node->var = func_match(node->arg1);
  else if (!strcmp(funcName, "blend"))
    node->var = func_acam(node->arg1, "blend");
  else if (!strcmp(funcName, "retrieve"))
    node->var = func_acam(node->arg1, "retrieve");
  else if (!strcmp(funcName, "orthogonalize"))
    node->var = func_orthogonalize(node->arg1);
  else if (!strcmp(funcName, "bitvec"))
    node->var = func_bitvec(node->arg1);
  else if (!strcmp(funcName, "rand"))
    node->var = func_rand(node->arg1);
  else if (!strcmp(funcName, "uniform"))
    node->var = func_uniform(node->arg1);
  else if (!strcmp(funcName, "randSeed"))
    node->var = func_randSeed(node->arg1);
  else if (!strcmp(funcName, "time"))
    node->var = func_time(node->arg1);
  else if (!strcmp(funcName, "randvec"))
    node->var = func_randVec(node->arg1);
  else if (!strcmp(funcName, "push"))
    node->var = func_push(node->arg1);
  else if (!strcmp(funcName, "pop"))
    node->var = func_pop(node->arg1);
  else if (!strcmp(funcName, "top"))
    node->var = func_top(node->arg1);
  else if (cadrString(funcName))
    node->var = func_cadr(funcName, node);
  else if (!strcmp(funcName, "cons"))
    node->var = func_cons(node->arg1);
  else if (!strcmp(funcName, "repeat"))
    node->var = func_repeat(node->arg1);
  else if (!strcmp(funcName, "print"))
    node->var = func_print(node->arg1);
  else if (!strcmp(funcName, "sqrt"))
    node->var = func_sqrt(node->arg1);
  else if (!strcmp(funcName, "normalize"))
    node->var = func_normalize(node->arg1);
  else if (!strcmp(funcName, "varalias"))
    node->var = func_varAlias(node->arg1);
  else if (!strcmp(funcName, "cos"))
    node->var = func_cosine(node->arg1);
  else
    AbortExpr("don't know how to evaluate function \"%s\"", funcName);

  HashTableDeleteByContext(ht, context);
  local_context--;
}

int getStoreItem(struct VarNode *item,
		 real **pvec,
		 char **pname,
		 bitWord **pbitVec,
		 int *pdim)
{
  if (item==NULL)
    return 0;
  if (item->type==LNode && item->d.l.car!=NULL && item->d.l.car->type==Vector
      && item->d.l.cdr!=NULL && item->d.l.cdr->type==LNode) {
    if (item->d.l.cdr->d.l.car==NULL) {
      if (pname!=NULL) *pname = item->d.l.car->name;
    } else if (item->d.l.cdr->d.l.car->type==String) {
      if (pname!=NULL) *pname = item->d.l.cdr->d.l.car->d.str;
    } else {
      return 0;
    }
    if (item->d.l.cdr->d.l.cdr==NULL) {
      if (pbitVec!=NULL) *pbitVec = NULL;
    } else if (item->d.l.cdr->d.l.cdr->type==BitVector) {
      if (pbitVec!=NULL) *pbitVec = item->d.l.cdr->d.l.cdr->d.vec.bw;
    } else {
      return 0;
    }
    if (pdim!=NULL) *pdim = item->d.l.car->d.vec.dim;
    if (pvec!=NULL) *pvec = item->d.l.car->d.vec.ptr;
    return 1;
  } else
    return 0;
}

/***
 * store(listvar, vector ...)
 * returns the value of listvar
 * the listvar is a chain, with each car a 2 node structure:
 *                    O
 *                   / \__________
 *                  O             \
 *                 / \             O    
 *              vec   O           / \______ ...
 *                   / \         O          
 *               name   bitvec  / \
 *			     vec   O       
 *			          / \      
 *			      name   bitvec
 *
 *
 */
struct VarNode *func_store(struct ExprNode *node)
{
  struct VarNode *mem = NULL;
  struct VarNode *vec, *end, *newItem, *vecName, *suppInfo, *bitVec = NULL;
  struct ExprNode *arg;
  int nargs = countListArgs(node);
  int storeBitVectors = trueString(stringVarValue("store.bitvectors"));
  if (nargs==0)
    AbortExpr("store: need at least one argument");
  else {
    int dim = 0;
    if (node->arg1->var == NULL)
      node->arg1->var_image = newLNodeVar(NULL, NULL);
    evalTree(node->arg1);		/* should give a list var */
    if (node->arg1->var_image != NULL) {
      freeVar(node->arg1->var_image);
      node->arg1->var_image = NULL;
    }
    mem = node->arg1->var;
    if (mem->type != LNode)
      AbortExpr("store: first argument must be an LNode");
    arg = node->arg2;
    for (end=mem; end->d.l.cdr!=NULL; end = end->d.l.cdr) {
      if (end->type!=LNode)
	AbortExpr("store: bad list structure");
      if (end->d.l.car!=NULL
	  && !getStoreItem(end->d.l.car, NULL, NULL, NULL, &dim))
	AbortExpr("store: bad item structure");
    }
    vecName = NULL;
    /* at the end, the cdr is nil, the car is may be used */
    while (arg!=NULL) {
      if (arg->type!=List)
	AbortExpr("store: bad argument list structure");
      evalTree(arg->arg1);
      if (arg->arg1->var==NULL)
	AbortExpr("store: argument evaluated to NULL!");
      else if (arg->arg1->var->type==String) {
	if (vecName!=NULL)
	  AbortExpr("store: two names in a row");
	vecName = copyVar(arg->arg1->var); /* remember for the next iter */
	vecName->permanent = 1;
      } else if (arg->arg1->var->type!=Vector) {
	AbortExpr("store: arguments must be vectors or strings");
      } else {
	vec = arg->arg1->var;
	if (dim==0)
	  dim = vec->d.vec.dim;
	else if (vec->d.vec.dim!=dim)
	  AbortExpr("store: new vector is different dimension (%d) to others (%d)", vec->d.vec.dim, dim);
	/* fprintf(dout, "store: adding %s\n", vec->name==NULL ? "(no name)" : vec->name); */
	if (vecName==NULL) {
	  if (!vec->permanent) vec->permanent = 1;
	} else  /* if we have a name, make a copy of the vector */
	  vec = copyVar(vec);
	if (storeBitVectors) {
	  bitVec = newBitVectorFromVector(vec);
	  bitVec->permanent = 1;
	}
	suppInfo = newLNodeVar(vecName, bitVec);
	suppInfo->permanent = 1;
	newItem = newLNodeVar(vec, suppInfo);
	newItem->permanent = 1;
	if (end->d.l.car==NULL)
	  end->d.l.car = newItem; /* when it's the first in the list */
	else {
	  end->d.l.cdr = newLNodeVar(newItem, NULL);
	  end = end->d.l.cdr;
	}
	vecName = NULL;
      }
      /* and go onto the next one */
      arg = arg->arg2;
    }
  }
  return mem;
}

/***
 * Returns a vector of pointers to items in the store
 * (see doc on func_store for structure of items)
 */
void findBestMatches(struct VarNode *mem, struct VarNode *probe,
		    enum SIM_TYPE similarity,
		    int capacity,
		    int useBitVectors,
		    real matchSpread,
		    struct VarNode **bestVar,
		    real *bestVal,
		    char *what)
{
  real matchVal = 0;
  real bitMatchVal = 0;
  real worstMatchValInBest = 0;
  real worstBitValInBest = 0;
  real spread = 0;
  int nBest = 0;
  int isStandardVectorType = standardVectorType(probe->d.vec.dim);
  struct VarNode *test, *probeBitVar;
  real *testVec;
  char *testName;
  bitWord *testBits, *probeBits = NULL;
  real *bestBitVal = NULL;
  int testDim;
  int i, j;
  int totalCount, skipCount;
  /* fprintf(dout, "matchType=%s dotprod=%d\n", matchType, dotprod); */
  /* fprintf(dout, "capacity=%d\n", capacity); */
  /* fprintf(dout, "probe=%s\n", probe->name); */
  /* fprintf(dout, "mem=%s\n", mem->name); */

  /* go through the list of vectors in mem */
  if (!isStandardVectorType && similarity!=DOTPRODUCT)
    AbortExpr("must use %s.sim = \"dotProduct\" with circular model", what);
  if (mem->d.l.car==NULL)
    AbortExpr("match: no vectors in memory");
  if (useBitVectors) {
    probeBitVar = newBitVectorFromVector(probe);
    freeVar(probeBitVar); /* pre-freeing: doesn't get freed til later */
    probeBits = probeBitVar->d.vec.bw;
    bestBitVal = CallocOrAbort(real, capacity);
    addToFreeList(bestBitVal); /* pre-freeing: doesn't get freed til later */
    /*
     * spread is the total expected range of bit-wise dot products
     * for any given real dot product.  It should probably vary
     * over the range, but to keep things simple, use a constant
     * value.  matchSpread is the number of std. devs on either
     * side of the mean, default value is 4
     */
    spread = 1.5*matchSpread*exp(-0.505*log((real) probe->d.vec.dim));
  }
  for (i=0; i<capacity; i++)
    bestVar[i] = NULL;
  totalCount = 0;
  skipCount = 0;
  for (; mem!=NULL; mem = mem->d.l.cdr) {
    test = mem->d.l.car;
    totalCount++;
    if (!getStoreItem(test, &testVec, &testName, &testBits, &testDim))
      AbortExpr("match: bad item structure in vector memory");
    if (testDim!=probe->d.vec.dim)
      AbortExpr("match: probe vector dim (%d) != memory vector dim (%d)",
		probe->d.vec.dim, testDim);

    if (useBitVectors) {
      if (testBits!=NULL)
	bitMatchVal = bitDotProduct(testDim, probeBits, testBits)*2.0/testDim - 1.0;
       else
	 bitMatchVal = 2.0;
      if (nBest>=capacity && bitMatchVal < worstBitValInBest-spread) {
	/* looks so bad don't try it */
	/* printf("Skipping %s - bitMatch was %g\n", testName, bitMatchVal); */
	skipCount++;
	continue;
      }
    }
    if (!isStandardVectorType)
      matchVal = sumCosDiff(testDim, probe->d.vec.ptr, testVec);
    else if (similarity==DOTPRODUCT)
      matchVal = dotProduct(testDim, probe->d.vec.ptr, testVec);
    else
      matchVal = cosine(testDim, probe->d.vec.ptr, testVec);

    /* fprintf(dout, "match: comparing with %s: %g (%g)\n",
	    testName, matchVal, bitMatchVal); */

    if (nBest<capacity || matchVal>worstMatchValInBest) {
      /* want to store this in the record of the best matches */
      for (i=nBest; i>0; i--) /* find the match val it is better than */
	if (matchVal<bestVal[i-1]) break;
      if (i>=capacity)
	AbortExpr("internal error: i too big");
      /* move the stored matches down */
      j = capacity-1;
      if (j>nBest) j = nBest;
      for ( ; j>i; j--) {
	bestVal[j] = bestVal[j-1];
	bestVar[j] = bestVar[j-1];
	if (useBitVectors)
	  bestBitVal[j] = bestBitVal[j-1];
      }
      bestVar[i] = test;
      bestVal[i] = matchVal;
      if (useBitVectors)
	bestBitVal[i] = bitMatchVal;
      if (nBest<capacity)
	nBest++;
      worstMatchValInBest = bestVal[nBest-1];
      if (useBitVectors)
        worstBitValInBest = bestBitVal[nBest-1];
    }
  }
  if (useBitVectors)
    fprintf(dout, "Skipped %d out of %d vectors\n", skipCount, totalCount);
}

/***
 * match()
 * returns the vector from the list that most closely matches probe
 */
struct VarNode *func_match(struct ExprNode *arg)
{
  struct VarNode *var;
  struct VarNode *mem = NULL;
  struct VarNode *probe = NULL;
  struct VarNode **bestVar = NULL;
  real *bestVec;
  char *bestName;
  real *bestVal;
  int dim;
  int useBitVecs = trueString(stringVarValue("match.filter"));
  real matchSpread = scalarVarValue("match.spread");
  int capacity = scalarVarValue("match.capacity");
  enum SIM_TYPE similarity = simType(stringVarValue("match.sim"));
  int print = trueString(stringVarValue("match.print"));
  int i;

  int nargs = countListArgs(arg);
  if (nargs!=2)
    AbortExpr("match: require two arguments, one vector, one lnode");
  if (arg->type!=List)
    AbortExpr("match: bad argument list structure");
  evalTree(arg->arg1);
  var = arg->arg1->var;
  if (var==NULL) {
    AbortExpr("match: failed to evaluate first argument");
  } else if (var->type==Vector) {
    probe = var;
  } else if (var->type==LNode) {
    mem = var;
  }
  arg = arg->arg2;
  evalTree(arg->arg1);
  var = arg->arg1->var;
  if (var==NULL) {
    AbortExpr("match: failed to evaluate second argument");
  } else if (var->type==Vector) {
    probe = var;
  } else if (var->type==LNode) {
    mem = var;
  }
  if (mem==NULL || probe==NULL)
    AbortExpr("match: require two arguments, one vector, one lnode");
  
  if (capacity<1)
    AbortExpr("match: capacity must be at least 1");

  bestVar = CallocOrAbort(struct VarNode *, capacity);
  bestVal = CallocOrAbort(real, capacity);

  findBestMatches(mem, probe, similarity, capacity, useBitVecs, matchSpread,
		  bestVar, bestVal, "match");

  if (!getStoreItem(bestVar[0], &bestVec, NULL, NULL, &dim))
    AbortExpr("match: bad structure in store list");
  var = newVectorVar(dim, bestVec);
  for (i=0; i<var->d.vec.dim; i++)
    var->d.vec.ptr[i] = bestVec[i];

  if (print) {
    int precision = (int) scalarVarValue("match.precision");
    for (i=0; i<capacity && bestVar[i]!=NULL; i++) {
      if (!getStoreItem(bestVar[i], NULL, &bestName, NULL, NULL))
	AbortExpr("match: bad structure in store list");
      if (bestName==NULL)
	bestName = "(no name)";
      fprintf(dout, "%-20s %.*g\n", bestName, precision, bestVal[i]);
    }
  }
  free(bestVar);
  free(bestVal);
  return var;
}

/***
 * blend(list, probe)
 * retrieve(list, probe)
 *
 * which="blend" or "retrieve"
 */
struct VarNode *func_acam(struct ExprNode *arg, char *what)
{
  struct VarNode *var;
  struct VarNode *mem = NULL;
  struct VarNode *probe = NULL;
  struct VarNode **bestVar = NULL;
  real *bestVec = NULL;
  char *name = NULL;
  int dim;
  real *bestVal;
  real **item = NULL;
  real weight = 0;
  int i, j;
  char buf[128];
  enum NORM_TYPE transfer;
  enum SIM_TYPE similarity;
  real transferBias;
  real transferThreshhold;
  real transferWeight;
  int capacity;

  int nargs = countListArgs(arg);

  sprintf(buf, "%s.transfer", what);
  transfer = normType(stringVarValue(buf));
  sprintf(buf, "%s.sim", what);
  similarity = simType(stringVarValue(buf));
  sprintf(buf, "%s.bias", what);
  transferBias = scalarVarValue(buf);
  sprintf(buf, "%s.threshhold", what);
  transferThreshhold = scalarVarValue(buf);
  sprintf(buf, "%s.weight", what);
  transferWeight = scalarVarValue(buf);
  sprintf(buf, "%s.capacity", what);
  capacity = scalarVarValue(buf);

  if (nargs!=2)
    AbortExpr("%s: require two arguments, one vector, one lnode", what);
  if (arg->type!=List)
    AbortExpr("%s: bad argument list structure", what);
  evalTree(arg->arg1);
  var = arg->arg1->var;
  if (var==NULL) {
    AbortExpr("%s: failed to evaluate first argument", what);
  } else if (var->type==Vector) {
    probe = var;
  } else if (var->type==LNode) {
    mem = var;
  }
  arg = arg->arg2;
  evalTree(arg->arg1);
  var = arg->arg1->var;
  if (var==NULL) {
    AbortExpr("%s: failed to evaluate second argument", what);
  } else if (var->type==Vector) {
    probe = var;
  } else if (var->type==LNode) {
    mem = var;
  }
  if (mem==NULL || probe==NULL)
    AbortExpr("%s: require two arguments, one vector, one lnode", what);

  if (capacity<1)
    AbortExpr("%s: capacity must be at least 1", what);

  bestVar = CallocOrAbort(struct VarNode *, capacity);
  bestVal = CallocOrAbort(real, capacity);
  item = CallocOrAbort(real *, capacity);

  findBestMatches(mem, probe, similarity, capacity, 0, 0.0,
		  bestVar, bestVal, what);

  if (!getStoreItem(bestVar[0], &bestVec, NULL, NULL, &dim))
    AbortExpr("%s: bad structure in store list", what);

  for (i=0; i<capacity && bestVar[i]!=NULL; i++) {
    if (!getStoreItem(bestVar[i], &bestVec, &name, NULL, NULL))
      AbortExpr("%s: bad structure in store list", what);

    if (transfer==LINEAR) {
      weight = bestVal[i];
    } else if (transfer==BOUNDEDLINEAR) {
      if (bestVal[i]<0)
	weight = 0;
      else if (bestVal[i]>1)
	weight = 1;
      else
	weight = bestVal[i];
    } else if (transfer==TANH) {
      weight = tanh(transferBias+transferWeight*bestVal[i]);
    } else if (transfer==LOGISTIC) {
      weight = 1.0/(1.0+exp(-(transferBias+transferWeight*bestVal[i])));
    } else if (transfer==HEAVISIDE) {
      weight = (bestVal[i] >= transferThreshhold) ? 1.0 : 0.0;
    } else {
      sprintf(buf, "%s.transfer", what);
      AbortExpr("%s: cannot use transfer function \"%s\"", what,
		stringVarValue(buf));
    }

    if (name==NULL) name = "(no name)";
    if (trueString(stringVarValue("debug.mem")))
      fprintf(dout, "Blending in %s with weight %g\n", name, weight);
    /* store weights in bestVal */
    bestVal[i] = weight;
    item[i] = bestVec;
  }
  capacity = i;

  var = newVectorVar(dim, bestVec);
  if (standardVectorType(var->d.vec.dim)) {
    /* standard vectors */
    for (j=0; j<var->d.vec.dim; j++)
      var->d.vec.ptr[j] = 0.0;
    for (i=0; i<capacity; i++)
      if (bestVal[i]!=0)
	for (j=0; j<var->d.vec.dim; j++)
	  var->d.vec.ptr[j] += bestVal[i] * item[i][j];
  } else {
    /* circular vectors */
    circularCOG(var->d.vec.dim, capacity, NULL, bestVal, item, var->d.vec.ptr);
  }

  free(bestVar);
  free(bestVal);
  free(item);
  return var;
}

/***
 * randseed()
 */
struct VarNode *func_randSeed(struct ExprNode *node)
{
  Real seed = scalarVarValue("rand.seed") + 1;
  struct VarNode *var = NULL;
  int nargs = countListArgs(node);
  if (nargs>1)
    AbortExpr("randseed: can take just zero or one arguments");
  if (nargs==1) {
    evalTree(node->arg1);		/* should give a scalar */
    var = node->arg1->var;
    if (var->type != Scalar)
      AbortExpr("randseed: argument must be a scalar");
    seed = var->d.value;
  }
  srand((int) seed);
  srandom(rand());
  setScalarVar("rand.seed", seed);
  var = newScalarVar(seed);
  return scalarVar("rand.seed");
}

/***
 * rand()
 */
struct VarNode *func_rand(struct ExprNode *arg)
{
  struct VarNode *var = NULL;
  int nargs = countListArgs(arg);
  real std_dev;
  real fac, r, v1, v2;
  int i;
  if (nargs>0) {
    while (arg!=NULL) {
      if (arg->type!=List)
	AbortExpr("rand: bad argument list structure");
      evalTree(arg->arg1);
      var = arg->arg1->var;
      if (var==NULL)
	AbortExpr("rand: failed to evaluate argument");
      if (var->type==Vector) {
	std_dev = sqrt(1.0/var->d.vec.dim);
	for (i=0; i+1<var->d.vec.dim; i+=2) {
	  do {
	    v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
	    v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
	    r=v1*v1+v2*v2;
	  } while (r >= 1.0);
	  fac = sqrt(-2.0*log(r)/r);
	  var->d.vec.ptr[i] = v2*fac*std_dev;
	  var->d.vec.ptr[i+1] = v1*fac*std_dev;
	}
	if (i<var->d.vec.dim) {
	  do {
	    v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
	    v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
	    r=v1*v1+v2*v2;
	  } while (r >= 1.0);
	  fac = sqrt(-2.0*log(r)/r);
	  var->d.vec.ptr[i] = v2*fac*std_dev;
	}
      } else if (var->type==Scalar) {
	std_dev = 1.0;
	do {
	  v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
	  v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
	  r=v1*v1+v2*v2;
	} while (r >= 1.0);
	fac = sqrt(-2.0*log(r)/r);
	var->d.value = v2*fac*std_dev;
      } else {
	AbortExpr("rand: args must be scalars or vectors");
      }
      /* and go onto the next one */
      arg = arg->arg2;
    }
  } else {
    std_dev = 1.0;
    do {
      v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
      v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
      r=v1*v1+v2*v2;
    } while (r >= 1.0);
    fac = sqrt(-2.0*log(r)/r);
    var = newScalarVar(v2*fac*std_dev);
  }
  return var;
}

/***
 * uniform()
 */
struct VarNode *func_uniform(struct ExprNode *arg)
{
  struct VarNode *var = NULL;
  int nargs = countListArgs(arg);
  if (nargs>0)
    AbortExpr("uniform: no arguments");
  var = newScalarVar(random()/(real)MAX_RAND_INT);
  return var;
}

/***
 * time()
 */
struct VarNode *func_time(struct ExprNode *arg)
{
  struct VarNode *var = NULL;
  int nargs = countListArgs(arg);
  static unsigned int current_time;
  static real wrapped_time = 0;
  real t;
  if (nargs>0)
    ErrorAbort("time: takes no arguments");
  current_time = (unsigned int) clock();
#ifdef sgi
  /* must have wrapped */
  if (current_time < last_time)
    wrapped_time += 2*2147;
  last_time = current_time;
#endif
  t = current_time; /* clock is in micro-seconds */
  var = newScalarVar((t+wrapped_time)*1e-6);
  return var;
}

void randomizeVec(struct VarNode *var)
{
  real std_dev;
  real fac, r, v1, v2;
  int i;
  if (standardVectorType(var->d.vec.dim)) {
    /* elements randomly distributed as N(0,1/n) */
    std_dev = sqrt(1.0/var->d.vec.dim);
    for (i=0; i+1<var->d.vec.dim; i+=2) {
      do {
	v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
	v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
	r=v1*v1+v2*v2;
      } while (r >= 1.0);
      fac = sqrt(-2.0*log(r)/r);
      var->d.vec.ptr[i] = v2*fac*std_dev;
      var->d.vec.ptr[i+1] = v1*fac*std_dev;
    }
    if (i<var->d.vec.dim) {
      do {
	v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
	v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
	r=v1*v1+v2*v2;
      } while (r >= 1.0);
      fac = sqrt(-2.0*log(r)/r);
      var->d.vec.ptr[i] = v2*fac*std_dev;
    }
  } else {
    /* elements are (r,theta) pairs, set r=1, theta in [-Pi,Pi] */
    if (var->d.vec.dim % 2 != 0)
      AbortExpr("randvec under circular model requires even dim");
    /* var->d.vec.ptr[0] = 1; */ /* first elt is (1,0) */
    /* var->d.vec.ptr[1] = 0; */
    for (i=0; i+1<var->d.vec.dim; i+=2) {
      var->d.vec.ptr[i] = 1;
      var->d.vec.ptr[i+1] = (random() / (MAX_RAND_INT/(2*M_PI)))-M_PI;
    }
  }
}

/***
 * randvec()
 */
struct VarNode *func_randVec(struct ExprNode *arg)
{
  struct VarNode *var = NULL;
  struct VarNode *var0 = newVectorVar(0, NULL);
  struct VarNode *lastVar = NULL;
  real x;
  int normalize = trueString(stringVarValue("randvec.normalize"));
  int dim = scalarVarValue("randvec.dim");

  freeVar(var0); /* pre-freeing: doesn't actually get freed now */

  for ( ; arg!=NULL; arg = arg->arg2) {
    if (arg->type!=List)
      AbortExpr("randvec: bad argument list structure");
    arg->arg1->var_image = var0;
    evalTree(arg->arg1);
    if (arg->arg1->var==NULL)
      AbortExpr("randvec: failed to evaluate argument");
    if (arg->arg1->var->type==Vector) {
      var = arg->arg1->var;
      if (var->d.vec.dim==0) {
	if (dim>0) {
	  var->d.vec.dim = dim;
	  var->d.vec.ptr = CallocOrAbort(real, dim);
	} else {
	  AbortExpr("randvec: dim <= 0 (!)");
	}
      }
      randomizeVec(var);
      if (normalize)
	normalizeVec(var->d.vec.dim, var->d.vec.ptr, NULL);
      lastVar = var;
    } else if (arg->arg1->var->type==Scalar) {
      x = arg->arg1->var->d.value;
      if (x<1 || floor(x)!=x)
	AbortExpr("randvec: dimension argument (%g) must be a positive integer", x);
      dim = floor(x);
    } else {
      AbortExpr("randvec: arguments must be integers or vectors");
    }
  }

  /* if there were no vector arguments, then make a random vector */
  if (lastVar==NULL) {
    if (dim>0) {
      var = newVectorVar(dim, NULL);
      randomizeVec(var);
      if (normalize)
	normalizeVec(var->d.vec.dim, var->d.vec.ptr, NULL);
      lastVar = var;
    } else {
      AbortExpr("randvec: dim <= 0 (!)");
    }
  }

  return lastVar;
}

/***
 * bitvec()
 */
struct VarNode *func_bitvec(struct ExprNode *arg)
{
  struct VarNode *arg1 = NULL;
  struct VarNode *var = NULL;
  int nargs = countListArgs(arg);
  if (nargs!=1)
    AbortExpr("bitvec: must have only one argument");
  if (!standardVectorType(0))
    AbortExpr("bitvec: not implemented under circular vector type");
  if (nargs==1) {
    evalTree(arg->arg1);
    arg1 = arg->arg1->var;
    if (arg1==NULL)
      AbortExpr("bitvec: failed to evaluate argument");
    if (arg1->type!=Vector)
      AbortExpr("bitvec: argument must be a vector");
  }
  var = newBitVectorFromVector(arg1);
  return var;
}

/***
 * func_orthogonalize(listvar)
 * returns an orthogonalized version of listvar
 *   listvar is an LNode
 */
struct VarNode *func_orthogonalize(struct ExprNode *arg)
{
  struct VarNode *mem = NULL;
  struct VarNode *orthmem = NULL;
  real *vec1, *vec2;
  struct VarNode *mp1, *mp2;
  Real c, v1len;
  int nargs = countListArgs(arg);
  int dim=0, prevdim = 0;
  int i;
  if (nargs!=1)
    AbortExpr("orthogonalize: need one argument");
  if (!standardVectorType(0))
    AbortExpr("orthogonalize: not implemented under circular vector type");

  if (arg->arg1->var == NULL)
    arg->arg1->var_image = newLNodeVar(NULL, NULL);
  evalTree(arg->arg1);		/* should give a list var */
  if (arg->arg1->var_image != NULL) {
    freeVar(arg->arg1->var_image);
    arg->arg1->var_image = NULL;
  }
  mem = arg->arg1->var;
  if (mem->type != LNode)
    AbortExpr("orthogonalize: first argument must be an LNode");
  /* run through the list and check it is all ok */
  for (mp1=mem; mp1!=NULL; mp1 = mp1->d.l.cdr) {
    if (!getStoreItem(mp1->d.l.car, NULL, NULL, NULL, &dim))
      AbortExpr("orthogonalize: bad item structure in store list");
    if (prevdim!=0 && dim!=prevdim)
      AbortExpr("orthogonalize: all vectors in store must be same dimension");
    if (mp1->type!=LNode)
      AbortExpr("orthogonalize: bad structure in store list");
    prevdim = dim;
  }
  /* remove components of initial vectors from subsequent ones */
  orthmem = copyVar(mem);
  for (mp1=orthmem; mp1->d.l.cdr!=NULL; mp1 = mp1->d.l.cdr) {
    if (!getStoreItem(mp1->d.l.car, &vec1, NULL, NULL, &dim))
      AbortExpr("orthogonalize: bad item structure in store list");
    v1len = dotProduct(dim, vec1, vec1);
    for (mp2=mp1->d.l.cdr; mp2!=NULL; mp2 = mp2->d.l.cdr) {
      if (!getStoreItem(mp2->d.l.car, &vec2, NULL, NULL, NULL))
	AbortExpr("orthogonalize: bad item structure in store list");
      c = dotProduct(dim, vec1, vec2)/v1len;
      /* printf("orthogonalizing vecs %x (%d) %x (%d) %g\n",
	     vec1, dim, vec2, dim, c); */
      for (i=0; i<dim; i++)
	vec2[i] -= c * vec1[i];
    }
  }
  /* make all the vectors have unit length */
  for (mp1=orthmem; mp1->d.l.cdr!=NULL; mp1 = mp1->d.l.cdr) {
    if (!getStoreItem(mp1->d.l.car, &vec1, NULL, NULL, &dim))
      AbortExpr("orthogonalize: bad item structure in store list");
    v1len = dotProduct(dim, vec1, vec1);
    if (v1len < scalarVarValue("orthogonalize.precision"))
      for (i=0; i<dim; i++)
	vec1[i] = 0.0;
    else {
      v1len = sqrt(v1len);
      for (i=0; i<dim; i++)
	vec1[i] /= v1len;
    }
  }
  return orthmem;
}

struct VarNode *func_push(struct ExprNode *node)
{
  struct VarNode *stack;
  struct VarNode *var;
  struct VarNode *new;
  struct ExprNode *arg;

  int nargs = countListArgs(node);
  if (nargs<1)
    AbortExpr("push: need at least one argument");
  if (node->arg1->var==NULL)
    node->arg1->var_image = newLNodeVar(NULL, NULL);
  evalTree(node->arg1);		/* should give a list var */
  if (node->arg1->var_image!=NULL) {
    freeVar(node->arg1->var_image);
    node->arg1->var_image = NULL;
  }
  stack = node->arg1->var;
  if (stack->type!=LNode)
    AbortExpr("push: first argument must be an LNode");
  arg = node->arg2;
  while (arg!=NULL) {
    if (arg->type!=List)
      AbortExpr("push: bad argument list structure");
    evalTree(arg->arg1);
    var = arg->arg1->var;
    if (var==NULL)
      AbortExpr("push: failed to evaluate argument");
    new = newLNodeVar(stack->d.l.car, stack->d.l.cdr);
    stack->d.l.cdr = new;
    stack->d.l.car = copyVar(var);
    /* and go onto the next one */
    arg = arg->arg2;
  }
  return stack;
}

struct VarNode *func_pop(struct ExprNode *node)
{
  struct VarNode *stack;
  struct VarNode *old;

  int nargs = countListArgs(node);
  if (nargs!=1)
    AbortExpr("pop: need one argument");
  evalTree(node->arg1);		/* should give a list var */
  stack = node->arg1->var;
  if (stack->type!=LNode)
    AbortExpr("pop: argument must be an LNode");
  if (stack->d.l.car!=NULL && !stack->d.l.car->permanent)
    freeVar(stack->d.l.car);
  stack->d.l.car = NULL;
  old = stack->d.l.cdr;
  if (old!=NULL) {
    stack->d.l.car = old->d.l.car;
    stack->d.l.cdr = old->d.l.cdr;
    old->d.l.car = old->d.l.cdr = NULL;
    freeVar(old);
  }
  return stack;
}

struct VarNode *func_top(struct ExprNode *node)
{
  struct VarNode *stack;

  int nargs = countListArgs(node);
  if (nargs!=1)
    AbortExpr("top: need one argument");
  evalTree(node->arg1);		/* should give a list var */
  stack = node->arg1->var;
  if (stack->type!=LNode)
    AbortExpr("top: argument must be an LNode");
  return stack->d.l.car;
}

int cadrString(char *str)
{
  if (str==NULL)
    return 0;
  if (*str=='c') {
    str++;
    while (*str=='a' || *str=='d')
      str++;
    if (*str=='r') {
      str++;
      if (*str=='\0')
	return 1;
    }
  }
  return 0;
}

struct VarNode *func_cadr(char *what, struct ExprNode *node)
{
  struct VarNode *cell, *new;
  int nargs = countListArgs(node);
  int i;
  if (nargs!=1)
    AbortExpr("%s: need one argument", what);
  evalTree(node->arg1);		/* should give a list var */
  if (node->arg1->var==NULL)
    AbortExpr("%s: can't follow NULL variable", what);
  if (node->arg1->var->type!=LNode)
    AbortExpr("%s: argument must be an LNode", what);
  cell = node->arg1->var;
  for (i=1; what[i]=='a' || what[i]=='d'; i++) {
    if (what[i]=='a') {
      if (cell->d.l.cdr!=NULL && !cell->d.l.cdr->permanent) {
	freeVar(cell->d.l.cdr);
	cell->d.l.cdr = NULL;
      }
      new = cell->d.l.car;
      if (!cell->permanent) {
	cell->d.l.car = NULL;
	freeVar(cell);
      }
      cell = new;
    } else {
      if (cell->d.l.car!=NULL && !cell->d.l.car->permanent) {
	freeVar(cell->d.l.car);
	cell->d.l.car = NULL;
      }
      new = cell->d.l.cdr;
      if (!cell->permanent) {
	cell->d.l.cdr = NULL;
	freeVar(cell);
      }
      cell = new;
    }
  }
  return cell;
}

struct VarNode *func_cons(struct ExprNode *node)
{
  struct VarNode *cell = NULL;
  int nargs = countListArgs(node);
  if (nargs!=1)
    AbortExpr("cons: need one argument");
  evalTree(node->arg1);		/* should give a list var */
  if (node->arg1->var==NULL)
    AbortExpr("cons: can't follow NULL variable");
  if (node->arg1->var->type!=LNode)
    AbortExpr("cons: argument must be an LNode");
  AbortExpr("cons: nyi");
  return cell;
}

struct VarNode *func_cosine(struct ExprNode *node)
{
  struct ExprNode *arg1;
  struct ExprNode *arg2;
  int n1, n2;
  real *rp1, *rp2, result;

  int nargs = countListArgs(node);
  if (!standardVectorType(0))
    AbortExpr("cosine: not implemented under circular vector type");
  if (nargs!=2)
    AbortExpr("cosine requires two arguments");
  evalTree(node->arg1);
  arg1 = node->arg1;
  evalTree(node->arg2->arg1);
  arg2 = node->arg2->arg1;
  if (arg1->var->type!=Vector || arg2->var->type!=Vector)
    AbortExpr("cosine requires two vectors");
  rp1 = nodeRealDataPtr(arg1, &n1, "cosine", 1);
  rp2 = nodeRealDataPtr(arg2, &n2, "cosine", 2);
  if (n1!=n2)
    AbortExpr("cosine require same length vectors");
  result = cosine(n1, rp1, rp2);
  return newScalarVar(result);
}

struct VarNode *func_sqrt(struct ExprNode *node)
{
  struct ExprNode *arg1;
  int n1;
  real *rp1, result;

  int nargs = countListArgs(node);
  if (nargs!=1)
    AbortExpr("sqrt requires one arguments");
  evalTree(node->arg1);
  arg1 = node->arg1;
  if (arg1->var->type!=Scalar)
    AbortExpr("sqrt requires scalar");
  rp1 = nodeRealDataPtr(arg1, &n1, "sqrt", 1);
  if (n1!=1)
    AbortExpr("sqrt requires scalar");
  result = sqrt(rp1[0]);
  return newScalarVar(result);
}

struct VarNode *func_repeat(struct ExprNode *node)
{
  struct ExprNode *arg1;
  int i;

  int nargs = countListArgs(node);
  if (nargs!=2)
    AbortExpr("repeat requires two arguments");
  evalTree(node->arg1);
  arg1 = node->arg1;
  if (arg1->var->type!=Scalar)
    AbortExpr("first argument for repeat must be number");
  /* printf("repeating %g times\n", arg1->var->d.value); */
  for (i=0; i<arg1->var->d.value; i++) {
    if (i>0)
      markUnevaluated(node->arg2->arg1);
    evalTree(node->arg2->arg1);
  }

  return node->arg2->arg1->var;
}

struct VarNode *func_print(struct ExprNode *arg)
{
  int count = 0;
  struct VarNode *var = NULL;
  char *sep = stringVarValue("print.separator");

  while (arg!=NULL && arg->arg1 != NULL) {
    evalTree(arg->arg1);
    var = arg->arg1->var;
    if (count>0 && strlen(sep)>0)
      fprintf(dout, "%s", sep);
    printVarValue(var, (int) scalarVarValue("print.level"));
    arg = arg->arg2;
    count++;
  }
  if (trueString(stringVarValue("print.newline")))
    fprintf(dout, "\n");
  if (trueString(stringVarValue("print.flush")))
    fflush(stdout);
  return var;
}

struct VarNode *func_varAlias(struct ExprNode *node)
{
  struct VarNode *var;
  struct HashTableItem *item;
  /* first check there are two arguments */
  if (countListArgs(node)!=2)
    AbortExpr("varAlias: must have two arguments");

  if (!node->arg1->type==Variable)
    AbortExpr("varAlias: first argument must be a variable");

  if (!node->arg2->arg1->type==Variable)
    AbortExpr("varAlias: first argument must be a variable");

  /* evaluate the first */
  evalTree(node->arg1);
  var = node->arg1->var;

  node = node->arg2->arg1;
  /* check that the second doesn't exist */
  item = HashTableLookup(ht, node->name, var_context);
  if (item!=NULL)
    AbortExpr("varAlias: \"%s\" is already a variable name", node->name);
  
  /* put the second in the hash table, pointing to the same as the first */
  item = HashTableInsert(ht, node->name, var_context, var);
  if (item==NULL)
    AbortExpr("could not create object named \"%s\"", node->name);
  if (var->permanent!=1)
    var->permanent = 1;
  node->var = var;

  return var;
}

enum NORM_TYPE normType(char *typeStr)
{
  enum NORM_TYPE type = 0;
  int len = strlen(typeStr);
  if (!strncasecmp("logistic", typeStr, len)
      || !strncasecmp("sigmoid", typeStr, len))
    type = LOGISTIC;
  else if (!strncasecmp("tanh", typeStr, len))
    type = TANH;
  else if (!strncasecmp("binary", typeStr, len))
    type = HEAVISIDE;
  else if (!strncasecmp("symmetric", typeStr, len))
    type = SYMMETRIC;
  else if (!strncasecmp("nopnorm", typeStr, len))
    type = NOPNORM;
  else if (!strncasecmp("unit", typeStr, len))
    type = UNITNORM;
  else if (!strncasecmp("norm2", typeStr, len))
    type = UNITNORM;
  else if (!strncasecmp("unitbinary", typeStr, len))
    type = UNITBINARY;
  else if (!strncasecmp("power", typeStr, len))
    type = POWERNORM;
  else if (!strncasecmp("linear", typeStr, len))
    type = LINEAR;
  else if (!strncasecmp("boundedlinear", typeStr, len))
    type = BOUNDEDLINEAR;
  else
    AbortExpr("normalize type (\"%s\") must be one of \"sigmoid\", \"tanh\",\n\"binary\", \"symmetric\", \"nopnorm\", \"unit\", \"unitbinary\", \"power\"\n\"linear\", or \"boundedlinear\"", typeStr);
  return type;
}

enum SIM_TYPE simType(char *typeStr)
{
  enum SIM_TYPE type = 0;
  int len = strlen(typeStr);
  if (!strncasecmp("dotproduct", typeStr, len))
    type = DOTPRODUCT;
  else if (!strncasecmp("cosine", typeStr, len))
    type = COSINE;
  else
    AbortExpr("similarity type (\"%s\") must be one of \"dotproduct\" or \"cosine\"", typeStr);
  return type;
}

void normalizeVec(int n, real *x, real *result)
{
  enum NORM_TYPE type;
  real lo = 0, hi = 0;
  double sum = 0, h;
  int i = 0;
  char *typeStr = stringVarValue("normalize.type");
  real threshhold = scalarVarValue("normalize.threshhold");
  type = normType(typeStr);

  if (result==NULL)
    result = x;

  if (type==HEAVISIDE)  { lo = 0.0; hi = 1.0; }
  if (type==SYMMETRIC)  { lo = -1.0; hi = 1.0; }
  if (type==UNITBINARY) { hi = sqrt(1.0/n); lo = -hi; }
  if (standardVectorType(n)) {
    switch (type) {
    case LOGISTIC:
      for (i=0; i<n; i++)
	result[i] = 1/(1+exp(-x[i]));
      break;
    case TANH:
      for (i=0; i<n; i++)
	result[i] = tanh(x[i]);
      break;
    case HEAVISIDE:
    case UNITBINARY:
    case SYMMETRIC:
      for (i=0; i<n; i++)
	result[i] = (x[i]>=threshhold ? hi : lo);
      break;
    case UNITNORM:
      sum = vSize(n, x);
      if (sum == 0.0)
	AbortExpr("Cannot normalize a zero vector");
      for (i=0; i<n; i++)
	result[i] = x[i]/sum;
      break;
    case NOPNORM:
    case LINEAR:
      for (i=0; i<n; i++)
	result[i] = x[i];
      break;
    case BOUNDEDLINEAR:
      for (i=0; i<n; i++) {
	result[i] = x[i];
	if (result[i] < 0)
	  result[i] = 0;
	else if (result[i] > 1)
	  result[i] = 1;
      }
      break;
    case POWERNORM:
      if (!trueString(stringVarValue("conv.fft")))
	AbortExpr("can only use powernorm when conv.fft=true");
      for (i=n; i>0 && (i&1) == 0; i>>=1) {};
      if (i!=1)
	AbortExpr("vector dim must be a power of 2 for conv/corr by fft");
      if (n>4096)
	AbortExpr("vector dim too large for conv/corr by fft: ffts don't work with n>4096");

      for (i=0; i<n; i++)
	result[i] = x[i];
      rfft(result, n);
      /* normalize each of the complex elements to have mag 1
       * taking into account that the complex transform is stored
       * in a compressed form, with elt 0 and 1 being real */
      result[0] = result[1] = 1.0;
      for (i=2; i<n; i+=2) {
	h = hypot(result[i], result[i+1]);
	if (sum==0.0) {
	  result[i] = 1.0;
	  result[i+1] = 0.0;
	} else {
	  result[i] /= h;
	  result[i+1] /= h;
	}
      }
      rifft(result, n);
      break;
    }
  } else {
    /* circular vector type - vector is (r,theta) pairs */
    switch (type) {

    case UNITNORM:
    case POWERNORM:
      for (i=0; i<n; i+=2) {
	if (result[i]<0)
	  result[i+1] = x[i+1]+M_PI;
	else
	  result[i+1] = x[i+1];
	result[i] = 1;			/* set r=1 */
	if (result[i+1]>M_PI)
	  result[i+1] = fmod(result[i+1]+M_PI, 2*M_PI)-M_PI;
	else if (result[i+1]<-M_PI)
	  result[i+1] = -fmod(-result[i+1]+M_PI, 2*M_PI)+M_PI;
      }
      break;
    case NOPNORM:
    case LINEAR:
      for (i=0; i<n; i+=2) {
	/* at least get into standard form, r>0 theta in [-Pi,Pi] */
	result[i] = x[i];
	if (result[i]<0)
	  result[i+1] = x[i+1]+M_PI;
	else
	  result[i+1] = x[i+1];
	if (result[i+1]>M_PI)
	  result[i+1] = fmod(result[i+1]+M_PI, 2*M_PI)-M_PI;
	else if (result[i+1]<-M_PI)
	  result[i+1] = -fmod(-result[i+1]+M_PI, 2*M_PI)+M_PI;
      }
      break;
    case LOGISTIC:
    case TANH:
    case HEAVISIDE:
    case UNITBINARY:
    case SYMMETRIC:    case BOUNDEDLINEAR:
      AbortExpr("cannot use normalize.type = \"%s\" under circular vector type", typeStr);
      break;
    }
  }
}

struct VarNode *func_normalize(struct ExprNode *node)
{
  real *rpr, *rp2;
  int nr, n2;
  struct VarNode *var;
  if (countListArgs(node)!=1)
    AbortExpr("normalize: must have just one argument");

  evalTree(node->arg1);
  var = dupVar(node->arg1->var);
  rp2 = nodeRealDataPtr(node->arg1, &n2, "normalize", 1);
  rpr = varRealDataPtr(var, &nr, NULL, 0);
  normalizeVec(nr, rp2, rpr);
  return var;
}
